#!/usr/bin/env python3


from __future__         import print_function

import  subprocess
import  shlex
import  argparse
import  os
import  sys
import  threading
import  time
from    queue           import  Queue
import  json
import  inspect
import  datetime
import  pprint
import  platform
from    functools       import  partial

import  pfmisc
from    pfmisc.debug    import  debug

import  pudb

def synopsis(ab_shortOnly = False):
    str_scriptName  = os.path.basename(sys.argv[0])
    str_shortSynopsis = """

    NAME

        %s

    SYNOPSIS


        %s \\
                    -c|--command <CMD>              \\
                    --pipeline                      \\
                    --ssh <user@machine[:<port>]>   \\
                    --auid <apparentUserID>         \\
                    --jid <jobID>                   \\
                    --prettyprint                   \\
                    --jsonprint                     \\
                    --showStdOut                    \\
                    --showStdErr                    \\
                    --echoCmd                       \\
                    --eventLoop                     \\
                    --verbosity <verbosity>         \\

        Run <CMD> in a variety of ways.


    """ % (str_scriptName, str_scriptName)
    str_description = """

    DESCRIPTION

        '%s' is a wrapper about an underlying system shell that
        executes commands in that shell.

        Standard returns (stdout, stderr) and exit codes are captured
        and available to other processes.

    ARGS

        -c|--command <CMD>
        Command to execute

        --verbosity <verbosity>
        The verbosity. Set to "-1" for silent.

        --pipeline
        If specified, treat compound commands (i.e. command strings
        concatenated together with ';') as a pipeline. Track each
        individual command separately.

        --auid <apparentUserID>
        Set the apparent User ID. This is a field stored at the root level
        of a job data structure and is useful is a job is run as one user
        but is apparently run as another. In particular, for certain cluster
        jobs, all jobs might be run as a single "master" user, but need to
        retain some concept of the apparent user that actually scheduled or
        started the job.

        --jid <jobID>
        Set the jobID, which is a field stored at the root level of a job
        data structure.

        --ssh <user@machine[:port]>
        If specified, ssh as <user> to <machine>[:port]> and execute the <CMD>
        on remote host.

        --prettyprint
        If specified, pretty print internal job dictionary object at finish.

        --jsonprint
        If specified, json.dumps() internal job dictionary object at finish.

        --showStdOut
        If specified, print the stdout of each job executed.

        --showStdErr
        If specified, print the stderr of each job executed.

        --echoCmd
        If specified, echo the actual command string that crunner will execute
        on the underlying shell.

        --eventLoop
        If specified, turn on job start/end event processing.

    """ % str_scriptName
    if ab_shortOnly:
        return str_shortSynopsis
    else:
        return str_shortSynopsis + str_description

class crunner(object):
    """
        This family of functor classes provides a unified interface
        to running shell-commands in several contexts:

            - locally on the underlying OS
            - remote host via ssh
            - cluster via scheduler

        Internally, 'crunner' spawns two threads. The first, 'ctl' is a
        control thread that in turn spawns the actual 'exe' thread that
        calls the subprocess module on the passed command line. The 'ctl'
        thread joins the 'exe' thread and so "waits" on it to complete.

            parent execution
        ----->--O----->---------O----->-- ... ------->----- ... ---->------>
                |              /   ^                                   ^
                |  __call__   /    | (startEvent)                      | (endEvent)
                O---->--+-->--     |                                   |
                        |          |                                   |
                        |          |     t_ctl thread (join t_exe)     |
                        --->-------O---->------------>------ ...--->---O--- (end)
                                 ^ |                                   ^
                                 | |                                  /|
                                 | |  t_exe thread                   / |
                                 | --------O--------->------ ... ---O--+- (end)
                                 |         |                       /   |
                                 |         | subprocess           /    |
                                 \         ---------->------ ... -    /
                                  \         (each job)               /
                                   -----<------------<------ ... -<-


        The main caller, however, returns to the parent context after a
        short interval. In this manner, main execution can continue in the parent
        process. The parent can query the crunner on the state of spawned
        threads, and can provide the object with callbacks that are executed
        on each job's startEvent and endEvent.

    """

    def __init__(self, **kwargs):
        """
        Initial object, set control and other flags.
        :param kwargs:
        """

        self.b_shell            = True
        self.b_showStdOut       = True
        self.b_showStdErr       = True
        self.b_echoCmd          = False

        # Debugging
        self.verbosity          = 1

        # Toggle special handling of "compound" jobs
        self.l_cmd              = []
        self.b_splitCompound    = True
        self.b_syncMasterSlave  = True      # If True, sub-commands will pause
                                            # after complete and need to be
                                            # explicitly told to continue by
                                            # master. The master needs to monitor
                                            # internal state to know when a thread
                                            # is waiting to be told to continue.

        # Queues for communicating between threads
        self.queue_startEvent   = Queue()
        self.queue_endEvent     = Queue()

        # Queues for extracting specific information
        # Separate queues for <start> and <end> events need to be maintained
        self.queue_pid          = Queue()
        self.queueStart_pid     = Queue()
        self.queueEnd_pid       = Queue()

        self.queue_continue     = Queue()
        self.queue              = Queue()

        # Job info
        # Each job executed is stored in a list of
        # dictionaries.
        self.jobCount           = 0
        self.jobTotal           = -1
        self.d_job              = {}
        self.d_fjob             = {}
        self.b_synchronized     = False
        self.b_jobsAllDone      = False
        self.jid                = ''
        self.auid               = ''

        # Threads for job control and execution
        self.t_ctl              = None      # Controller thread
        self.t_exe              = None      # Exe thread

        self.pp                 = pprint.PrettyPrinter(indent=4)

        str_debugFile           = ''
        b_debugToFile           = False
        for key, val in kwargs.items():
            if key == 'verbosity':      self.verbosity  = val
            if key == 'debugFile':      str_debugFile   = val
            if key == 'debugToFile':    b_debugToFile   = val
        self.debug              = pfmisc.debug(
                                        verbosity       = self.verbosity,
                                        debugFile       = str_debugFile,
                                        debugToFile     = b_debugToFile,
                                        within          = 'crunner')

    def __call__(self, str_cmd, **kwargs):
        """
        The __call__ functor essentially creates a controller thread
        which in turns creates an exe thread for the subprocess call.

        The __call__ thus returns to the parent context with minimal
        delay.

        :param str_cmd:
        :param kwargs:
        :return:
        """

        if self.b_echoCmd: print(str_cmd)
        l_args          = shlex.split(str_cmd)
        self.l_cmd      = []
        if self.b_splitCompound:
            self.l_cmd  = str_cmd.split(";")
        else:
            self.l_cmd.append(str_cmd)

        self.t_ctl      = threading.Thread(target = self.ctl,
                                            args   = (str_cmd,),
                                            kwargs = kwargs)
        self.t_ctl.start()

        # We wait a small interval to give the exe thread enough
        # time to start the subprocess and determine its PID
        time.sleep(0.5)

    def ctl(self, str_cmd, **kwargs):
        """
        A controller thread.

        :param q:
        :return:
        """

        self.debug.qprint(msg = 'start ctl', level=2)

        self.jobTotal   = len(self.l_cmd)

        for job in self.l_cmd:
            # This thread runs independently of the master, so set
            # an unsynchronized flag
            self.b_synchronized         = False


            # Declare the structure
            str_jobCount    = str(self.jobCount)
            self.d_job[str_jobCount]   = {}
            self.d_fjob[str_jobCount]  = {}

            self.t_exe                  =   threading.Thread(   target = self.exe,
                                                                args   = (job,),
                                                                kwargs = kwargs)
            self.t_exe.start()
            self.t_exe.join()

            # self.d_job[str_jobCount]   = self.queue.get()

            # Now, "block" until the parent thread says we can continue...
            self.debug.qprint("waiting for continue event from master...", level=2)
            blockUntil = self.queue_continue.get()
            self.debug.qprint("continue event processed", level=2)

            # Increase the job count
            self.jobCount               += 1
            self.debug.qprint("increasing job count to %d" % self.jobCount, level = 2)

        self.debug.qprint("b_synchronized = %r" % self.b_synchronized, level=2)
        self.debug.qprint(msg = 'done ctl', level=2)
        self.b_jobsAllDone  = True

    def queue_pop(self, **kwargs):
        """
        get() [i.e. pop()] from a queue specified in the kwargs

        The main purpose of this method is to control the size of the queue in the
        main exe loop. Each timeout iteration a pid is written to a specific queue, and
        popped during an exception handling.

        Once the event is finished, the pid queues must ONLY have ONE entry, not multiple.

        :param kwargs:
        :return:
        """
        queue = None
        str_queue = ""
        for key, val in kwargs.items():
            if key == 'queue':  str_queue   = val

        if str_queue == 'queueStart':   queue = self.queueStart_pid
        if str_queue == 'queueEnd':     queue = self.queueEnd_pid

        if queue.qsize():
            self.debug("%s queue has %d elements" % (str_queue, queue.qsize()),
                       level=3)
            pid = queue.get()
            self.debug("%s queue popping pid %d..." % (str_queue, pid), level=3)
            self.debug("%s queue has %d elements"   % (str_queue, queue.qsize()),
                       level=3)

    def queue_flush(self, **kwargs):
        """
        Clear a queue.

        :param kwargs:
        :return:
        """
        queue = None
        str_queue = ""
        for key, val in kwargs.items():
            if key == 'queue':  str_queue   = val

        if str_queue == 'queueStart':   queue = self.queueStart_pid
        if str_queue == 'queueEnd':     queue = self.queueEnd_pid

        self.debug("Flushing %s (currently contains %d elements)" % (str_queue, queue.qsize()), level=3)
        for q in range(0, queue.qsize()):
            pid = queue.get()
        self.debug("Flushed %s (currently contains %d elements)" % (str_queue, queue.qsize()), level=3)

    def exe(self, str_cmd, **kwargs):
        """
        Execute <str_cmd> on the underlying shell in a separate thread.

        Job/process info is written to the thread queue construct

        :param str_cmd: command to execute
        :param kwargs:
        """

        b_ssh   = False
        wait    = 0.3

        for key,val in kwargs.items():
            if key == 'ssh':    b_ssh   = bool(val)
            if key == 'wait':   wait    = val

        self.debug.qprint(msg    = "start << %s >> (%d) " % (str_cmd, self.verbosity),
                         level  = 3)
        str_jobCount    = str(self.jobCount)
        self.d_job[str_jobCount]['cmd']                = str_cmd
        self.d_job[str_jobCount]['done']               = False
        self.d_job[str_jobCount]['started']            = True
        self.d_job[str_jobCount]['startTrigger']       = True
        self.d_job[str_jobCount]['eventFunctionDone']  = False
        if len(self.auid):
            self.d_job[str_jobCount]['auid']           = self.auid
        if len(self.jid):
            self.d_job[str_jobCount]['jid']            = self.jid

        # Platform info
        self.d_job[str_jobCount]['system']             = platform.system()
        self.d_job[str_jobCount]['machine']            = platform.machine()
        self.d_job[str_jobCount]['platform']           = platform.platform()
        self.d_job[str_jobCount]['uname']              = platform.uname()
        self.d_job[str_jobCount]['version']            = platform.version()

        if b_ssh:
            self.d_job[str_jobCount]['pid_remote']     = ""
            self.d_job[str_jobCount]['remoteHost']     = self.str_remoteHost
            self.d_job[str_jobCount]['remoteUser']     = self.str_remoteUser
            self.d_job[str_jobCount]['remotePort']     = self.str_remotePort
            self.d_job[str_jobCount]['sshInput']       = self.str_sshInput

        self.queue_flush(queue = 'queueStart')
        self.queue_flush(queue = 'queueEnd')

        self.queue_startEvent.put({'startTrigger': True})

        self.d_job[str_jobCount]['startstamp']         = '%s' % datetime.datetime.now()
        proc = subprocess.Popen(
            str_cmd,
            stdout              = subprocess.PIPE,
            stderr              = subprocess.PIPE,
            shell               = True,
            universal_newlines  = True
        )

        b_subprocessRunning     = True
        pollLoop                = 0
        self.remotePID          = ""
        while b_subprocessRunning:
            try:
                # self.queue_pid.put(proc.pid)
                self.queueStart_pid.put(proc.pid)
                self.queueEnd_pid.put(proc.pid)
                self.debug.qprint("Putting pid %d in queues..." % proc.pid, level=3)
                o = proc.communicate(timeout=0.1)
                b_subprocessRunning = False
            except subprocess.TimeoutExpired:
                if b_ssh:
                    self.remotePID    += proc.stdout.readline().strip()
                    self.d_job[str_jobCount]['pid_remote'] = self.remotePID
                # pid = self.queue_pid.get()
                if self.queueStart_pid.qsize(): self.queue_pop(queue = 'queueStart')
                if self.queueEnd_pid.qsize():   self.queue_pop(queue = 'queueEnd')
                self.d_job[str_jobCount]['pid']    = proc.pid
                pollLoop += 1
                if pollLoop % 10 == 0:
                    self.debug.qprint("job %d (%s) running... started = %r" %
                                     (self.jobCount,
                                      proc.pid,
                                      self.d_job[str_jobCount]['started']),
                                     level = 3)

        self.d_job[str_jobCount]['endstamp']       = '%s' % datetime.datetime.now()
        self.d_job[str_jobCount]['done']           = True
        self.d_job[str_jobCount]['doneTrigger']    = True
        self.d_job[str_jobCount]['started']        = False
        self.d_fjob[str_jobCount]['proc']          = proc
        self.d_job[str_jobCount]['pid']            = proc.pid
        self.d_job[str_jobCount]['returncode']     = proc.returncode
        self.d_job[str_jobCount]['stdout']         = o[0]
        self.d_job[str_jobCount]['stderr']         = o[1]
        self.d_job[str_jobCount]['cmd']            = str_cmd

        self.debug.qprint("Putting end event in queue...", level=3)
        self.queue_endEvent.put({'endTrigger': True})
        self.queue.put(self.d_job[str_jobCount])

        self.debug.qprint(msg = "done << %s >> " % str_cmd, level=3)

    def exe_stillRunning(self):
        """
        Checks if the exe thread is still running.
        """
        return self.t_exe.isAlive()

    def tag_print(self, **kwargs):
        """
        Simply print the passed tag from the internal data structure, but block
        possibly on error -- there are thread timing considerations and it is
        possible that the child thread hasn't started the job by the time this
        method attempts to access the tag.

        :param kwargs:
        :return:
        """

        timeout = 0.1
        str_tag = 'pid'

        for key, val in kwargs.items():
            if key == 'tag':        str_tag = val
            if key == 'timeout':    timeout = val

        pollLoop        = 0
        str_jobCount    = str(self.jobCount)
        b_tagPrinted    = False
        print("in tag_print...")
        while not b_tagPrinted:
            try:
                print("%s = %s" % (str_tag, self.d_job[str_jobCount][str_tag]))
                b_tagPrinted    = True
            except:
                pollLoop += 1
                if pollLoop % 10 == 0:
                    self.debug.qprint("tag not available")
                time.sleep(timeout)

    def pid_get(self, **kwargs):
        """
        Simply reads the pid from the queue_pid().
        """

        b_queueStart_pid    = False
        b_queueEnd_pid      = False
        str_queue           = ""

        for key, val in kwargs.items():
            if key == 'queue':  str_queue   = val

        queue = None
        if str_queue == "queueStart":   queue = self.queueStart_pid
        if str_queue == "queueEnd":     queue = self.queueEnd_pid

        self.debug.qprint("getting %s pid... (queue size = %d)" % (str_queue, queue.qsize()))
        pid = queue.get()
        self.debug.qprint("%s pid = %d" % (str_queue, pid))
        return pid

    def currentJob_waitUntilDone(self, **kwargs):
        """
        Waits until the current job is done. This method
        monitors both the self.b_currentJobDone flag and
        also checks that the exe thread is alive.

        :param kwargs:
        :return:
        """

        timeout     = 0.1

        for key,val in kwargs.items():
            if key == 'timeout':    timeout = val

        pollLoop    = 0
        str_jobCount    = str(self.jobCount)
        while not self.d_job[str_jobCount]['done'] and self.exe_stillRunning():
            pollLoop += 1
            if pollLoop % 10 == 0:
                self.debug.qprint(msg    = "job queue = %d / %d, -->b_synchronized = %r<--" %
                                          (self.jobCount,
                                           self.jobTotal,
                                           self.b_synchronized),
                                 level  = 0)
            time.sleep(timeout)

    def jobInfo_get(self, **kargs):
        """
        A buffered method to access elements of the d_job dictionary. Since
        the requested field might not exist when this function is called,
        the method blocks in a limited timeout.

        Only a certain number of timeouts are allowed before this method
        gives up with an error.

        :param kargs:
        :return:
        """

        timeout             = 0.1
        attemptsTotal       = 500
        job                 = self.jobCount
        field               = 'pid'

        for key, val in kargs.items():
            if key == 'job':            job             = val
            if key == 'field':          field           = val
            if key == 'timeout':        timeout         = val
            if key == 'attemptsTotal':  attemptsTotal   = val

        attempts    = 0
        b_success   = False
        ret         = None

        while not b_success and job < self.jobTotal:
            try:
                ret         = self.d_job[str(job)][field]
                b_success   = True
            except:
                time.sleep(timeout)
                attempts += 1
                if attempts > attemptsTotal:
                    b_success   = False
                    break

        return{'success':   b_success,
               'field':     ret}

    def job_done(self, **kwargs):
        """

        Simply returns a boolean if the current or (other) job is done.

        :param kwargs:
        :return:
        """

        job     = self.jobCount
        b_end   = False

        for k, v in kwargs.items():
            if k == 'job':  job = v

        self.debug.qprint("End Queue contains: %d events" % self.queue_endEvent.qsize())
        self.debug.qprint("End Queue current job = %d/%d" % (self.jobCount, self.jobTotal))
        if self.jobCount < self.jobTotal:
            b_end = self.queue_endEvent.get()
            self.debug.qprint("Got an end event!")
        return b_end

    def job_started(self, **kwargs):
        """

        Simply returns a boolean if the current or (other) job has started

        :param kwargs:
        :return:
        """

        job     = self.jobCount
        b_start = False

        for k, v in kwargs.items():
            if k == 'job':  job = v

        self.debug.qprint("Start Queue contains: %d events" % self.queue_startEvent.qsize())
        self.debug.qprint("Start Queue current job = %d/%d" % (self.jobCount, self.jobTotal))
        # pudb.set_trace()
        if self.jobCount < self.jobTotal:
            b_start = self.queue_startEvent.get()
            self.debug.qprint("Got a start event!")
        return b_start

    def jobs_loopctl(self, **kwargs):
        """
        This is an even loop entry point.

        The particular method can fire callbacks at two distinct epochs:

            * onJobStart
            * onJobDone

        These call back functions accept a standard arg construct and must return a dictionary
        of which one element contain a boolean 'status'.

        :param kwargs:
        :return:
        """

        timeout         = 0.1
        b_onJobStart    = False
        f_onJobStart    = None
        b_onJobDone     = False
        f_onJobDone     = None

        for key,val in kwargs.items():
            if key == 'timeout':        timeout         = val
            if key == 'onJobStart':
                f_onJobStart    = val
                b_onJobStart    = True
            if key == 'onJobDone':
                f_onJobDone     = val
                b_onJobDone     = True

        # This is the main event loop for this method. Each loop of the while
        # encompasses the lifetime of a single job. Trigger events are
        # evaluated first, and at the end of the loop, processing waits to make
        # sure the job is in fact done so that the main thread can communicate
        # to the ctl thread that the next job can be processed.
        while self.jobCount < self.jobTotal:
            # This short timeout during the very last job loop is needed for
            # synchronization at the very end of the processing loop. It's
            # probably a very very BAD idea to do this but I couldn't figure
            # out another way. The final loop needed a small delay at the start
            # so that the final all-loops-done event can sync properly.
            if self.jobCount == self.jobTotal-1: time.sleep(timeout)

            # Wait for the start event trigger
            while not self.job_started() and self.jobCount < self.jobTotal: pass
            if b_onJobStart and self.jobCount < self.jobTotal: f_onJobStart()

            # Finally, before looping on, we need to wait until the job is in fact
            # over and then set the synchronized flag to tell the ctl thread
            # it's safe to continue
            while not self.job_done() and self.jobCount < self.jobTotal: pass
            if b_onJobDone and self.jobCount < self.jobTotal: f_onJobDone()
            self.debug.qprint("Setting synchronized flag from %r to True" % self.b_synchronized)
            self.debug.qprint("jobCount / jobTotal = %d / %d" % (self.jobCount, self.jobTotal))
            self.b_synchronized     = True
            self.debug.qprint("continue queue has length %d" % self.queue_continue.qsize())
            self.debug.qprint("putting ok in continue queue")
            self.queue_continue.put(True)

    # This method is somewhat historical and largely depreciated. It is left here
    # for reference/legacy purposes.
    def jobs_eventLoop(self, **kwargs):
        """
        This an event loop entry point.

        Two event/condition families of callbacks should be provided
        by the caller:

            * a list of functions to execute on specific events in
              the queue:

                * onEventTriggered_do
                * onNewJobSpawned_do

            * a list of functions to evaluate when the event loop
              should terminate

        The caller can provide
        callback functions to be triggered on events in the
        proces7sing loop.

        A caller can provide a list of functions to execute on
        events in the job queue and should also provide a terminating
        condition for this event loop.

        Events include:

            * onEventTriggered_do
            * onNewJobSpawned_do


        :param kwargs:
        :return:
        """

        timeout                 = 0.1
        onEventTriggeredDo      = None
        b_onEventTriggeredDo    = False
        f_waitForEvent          = None

        for key,val in kwargs.items():
            if key == 'timeout':        timeout             = val
            if key == 'waitForEvent':   f_waitForEvent      = val
            if key == 'onEventTriggeredDo':
                onEventTriggeredDo    = val
                b_onEventTriggeredDo  = True

        # This is the main event loop for this method. Each loop of the while
        # encompasses the lifetime of a single job. Trigger events are
        # evaluated first, and at the end of the loop, processing waits to make
        # sure the job is in fact done so that the main thread can communicate
        # to the ctl thread that the next job can be processed.
        while self.jobCount < self.jobTotal:
            # Keep the currentJob id fixed during this loop -- while this loop
            # actually runs, the self.jobCount might change, but this loop will
            # only process the self.jobCount as captured here.
            currentJob  = self.jobCount

            time.sleep(timeout)

            # Wait for the event trigger
            self.debug.qprint("EL: CurrentJob = %d/%d" %(self.jobCount, self.jobTotal))
            while not eval(f_waitForEvent) and self.jobCount < self.jobTotal: pass
            if b_onEventTriggeredDo and self.jobCount < self.jobTotal:
                eval(onEventTriggeredDo)
                self.d_job[str(currentJob)]['eventFunctionDone']   = True
                self.debug.qprint(msg = "currentJob = %d job queue = %d / %d  -->b_synchronized = %r<--" %
                                       (currentJob,
                                        self.jobCount,
                                        self.jobTotal,
                                        self.b_synchronized), level = 0)

            # Finally, before looping on, we need to wait until the job is in fact
            # over and then set the synchronized flag to tell the ctl thread
            # it's safe to continue
            if self.queueEnd_pid.qsize():
                while not self.job_done() and self.jobCount < self.jobTotal: pass
            self.debug.qprint("Setting synchronized flag from %r to True" % self.b_synchronized)
            self.b_synchronized     = True
            self.debug.qprint("continue queue has length %d" % self.queue_continue.qsize())
            self.debug.qprint("putting ok in continue queue")
            self.queue_continue.put(True)

    def exe_waitUntilDone(self, **kwargs):
        """
        Blocks in method until thread is done, i.e. !isAlive()
        """
        timeout = 1

        for key,val in kwargs.items():
            if key == 'timeout':    timeout = val

        while self.exe_stillRunning():
            time.sleep(timeout)

    def exitOnDone(self, **kwargs):
        """
        An exit function that blocks on any threads that might still be
        running. If called with

            timeout     = timeout

        will sleep for timeout seconds between polling. If called with

            returncode  = code

        will force exit to system with passed code, otherwise will
        pass to system exitcode of spawned subprocess.
        """
        timeout             = 0.1
        b_overrideReturn    = False
        for key,val in kwargs.items():
            if key == 'timeout':    timeout = val
            if key == 'returncode':
                                    b_overrideReturn    = True
                                    returncode          = val
        self.exe_waitUntilDone(timeout = timeout)
        str_jobCount    = str(self.jobCount-1)
        if not b_overrideReturn:
            returncode  = self.d_job[str_jobCount]['returncode']

        if self.b_showStdOut:
            for j in range(0, self.jobTotal):
                print(self.d_job[str(j)]['stdout'])
        if self.b_showStdErr:
            for j in range(0, self.jobTotal):
                print(self.d_job[str(j)]['stderr'])

        sys.exit(returncode)

    def run(self, str_cmd, **kwargs):
        """
        A convenience function that "runs" the passed <str_cmd> and returns information
        of job "0". Different jobs can be specified by kwargs for compound statements.

        Returns a dictionary:

            {
                "stdout":   <stdout from job>,
                "stderr":   <stderr from job>,
                "exitCode": <exitCode from job>
            }

        """

        str_job = "0"

        for k,v in kwargs.items():
            if k == 'job':  str_job = v

        self.__call__(str_cmd)
        self.jobs_loopctl()

        d_ret                   = {}
        d_ret['stdout']         = self.d_job[str_job]['stdout']
        d_ret['stderr']         = self.d_job[str_job]['stderr']
        d_ret['returncode']     = self.d_job[str_job]['returncode']

        return d_ret

class crunner_ssh(crunner):
    """
    A specialized class that handles ssh connections to remote
    hosts.

    Effectively this prepends commands with <sshString> and captures
    the remote process ID.

    """

    def __init__(self, **kwargs):
        """
        The constructor for this class. The ssh argument from the
        command line is of form:

            username@host:port

        """

        crunner.__init__(self, **kwargs)

        self.str_remoteUser = ''
        self.str_remoteHost = ''
        self.str_remotePort = '22'
        self.str_sshInput   = ''
        self.remotePID      = ''

        for key, val in kwargs.items():
            if key == 'ssh':    str_sshInput    = val

        self.str_remoteUser = str_sshInput.split('@')[0]
        l_namePort          = str_sshInput.split(':')
        if len(l_namePort) > 1:
            self.str_remotePort = l_namePort[1]
        self.str_remoteHost = l_namePort[0].split('@')[1]

    def __call__(self, str_cmd, *args, **kwargs):
        """
        Pre and post-pend the call with ssh related constructs.

        If the <str_cmd> is a compound shell expression (i.e. has
        semi-colon ';' joins), then each component is executed
        separately.
        """

        self.l_cmd      = []
        if self.b_splitCompound:
            self.l_cmd  = str_cmd.split(";")
        else:
            self.l_cmd.append(str_cmd)

        i = 0
        for j in self.l_cmd:
            str_sshCmd = "ssh -p %s %s@%s \'%s & echo $!\'" % (
                self.str_remotePort,
                self.str_remoteUser,
                self.str_remoteHost,
                j
            )
            self.l_cmd[i] = str_sshCmd
            i += 1

        str_cmd = " ; ".join(self.l_cmd)

        kwargs['ssh']    = True
        crunner.__call__(self, str_cmd, **kwargs)
        self.waitForRemotePID()

    def waitForRemotePID(self, **kwargs):
        """
         This method blocks until the ssh connection has been made and
         the remote processes spawned and its PID captured.
        """
        timeout     = 0.1

        for key,val in kwargs.items():
            if key == 'timeout':    timeout = val

        while not len(self.remotePID):
            time.sleep(timeout)

def pid_print(shell, **kwargs):

    str_queue   = "None"
    b_silent    = False
    for k, v in kwargs.items():
        if k == 'queue':    str_queue   = v
        if k == 'silent':   b_silent    = v

    pid = shell.pid_get(**kwargs)
    if not b_silent: print("%20s: id of process: %d" % (str_queue, pid))

def job_started(shell):
    return shell.job_started()


"""
OpenShift cluster manager module that provides functionality to schedule jobs as well as
manage their state in the cluster.
"""

import yaml
import json
import os
from kubernetes import client as k_client, config
from kubernetes.client.rest import ApiException

class OpenShiftManager(object):

    def __init__(self, project=None):
        self.kube_client = None
        self.kube_v1_batch_client = None
        self.project = project or os.environ.get('OPENSHIFTMGR_PROJECT') or 'myproject'

        # init the openshift client
        self.init_openshift_client()

    def init_openshift_client(self):
        """
        Method to get a OpenShift client connected to remote or local OpenShift
        """
        kubecfg_path = os.environ.get('KUBECFG_PATH')
        if kubecfg_path is None:
            config.load_kube_config()
        else:
            config.load_kube_config(config_file=kubecfg_path)
        self.kube_client = k_client.CoreV1Api()
        self.kube_v1_batch_client = k_client.BatchV1Api()

    # The kubecfg secret being mounted in the third container (publish) is the one that is generated by the service account
    def schedule(self, image, command, name, number_of_workers, 
                 cpu_limit, memory_limit, gpu_limit, incoming_dir, outgoing_dir):
        """
        Schedule a new job and returns the job object.
        """
        d_job = {
            "apiVersion": "batch/v1",
            "kind": "Job",
            "metadata": {
                "name": name
            },
            "spec": {
                "ttlSecondsAfterFinished": 20,
                "parallelism": number_of_workers,
                "completions": number_of_workers,
                "activeDeadlineSeconds": 3600,
                "template": {
                    "metadata": {
                         "labels":{
                            "job-origin":"pman"
                        },
                        "name": name
                    },
                    "spec": {
                        "restartPolicy": "Never",
                        "containers": [
                            {
                                "env": [
                                    {
                                        "name": "NUMBER_OF_WORKERS",
                                        "value": number_of_workers
                                    },
                                    {
                                        "name": "CPU_LIMIT",
                                        "value": cpu_limit
                                    },
                                    {
                                        "name": "MEMORY_LIMIT",
                                        "value": memory_limit
                                    }
                                ],
                                "name": name,
                                "image": image,
                                "command": command.split(" "),
                                "resources": {
                                    "limits": {
                                        "memory": memory_limit,
                                        "cpu": cpu_limit
                                    },
                                    "requests": {
                                        "memory": "150Mi",
                                        "cpu": "250m"
                                    }
                                },
                                "volumeMounts": [
                                    {
                                        "mountPath": "/share",
                                        "name": "shared-volume"
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }
        if int(gpu_limit) > 0:  # Typecasting before a check
            # The assumption is containers[0] is always image plugin pod as the publish container is appended later.
            # These is specific to 3.9+  Ref: https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/
            d_job['spec']['template']['spec']['containers'][0]['resources']['limits']['nvidia.com/gpu'] = int(gpu_limit)
            # Add node selector for node.
            # d_job['spec']['template']['spec']['nodeSelector'] = {'accelerator': 'gpu-node'}
            d_job['spec']['template']['spec']['containers'][0]['securityContext']= {
                "allowPrivilegeEscalation": False,
						"capabilities": {
							"drop": [
								"ALL"
							]
						}
            }
            env = d_job['spec']['template']['spec']['containers'][0]['env']
            env.extend(({
							"name": "NVIDIA_VISIBLE_DEVICES",
							"value": "all"
						},
						{
							"name": "NVIDIA_DRIVER_CAPABILITIES",
							"value": "compute,utility"
						},
						{
							"name": "NVIDIA_REQUIRE_CUDA",
							"value": "cuda>=9.0"
						}))



        if os.environ.get('STORAGE_TYPE') == 'swift':
            d_job['spec']['template']['spec']['initContainers'] = [
                {
                    "name": "init-storage",
                    "image": "fnndsc/pman-swift-publisher",
                    "env": [
                        {
                            "name": "SWIFT_KEY",
                            "value": name
                        },
                        {
                            "name": "INCOMING_DIR",
                            "value": incoming_dir
                        },
                        {
                            "name": "OUTGOING_DIR",
                            "value": outgoing_dir
                        }
                    ],
                    "command": [
                        "python3",
                        "get_data.py"
                    ],
                    "resources": {
                        "limits": {
                            "memory": "1024Mi",
                            "cpu": "2000m"
                        },
                        "requests": {
                            "memory": "150Mi",
                            "cpu": "250m"
                        }
                    },
                    "volumeMounts": [
                        {
                            "mountPath": "/share",
                            "name": "shared-volume"
                        },
                        {
                            "mountPath": "/etc/swift",
                            "name": "swift-credentials",
                            "readOnly": True
                        },
                        {
                            "mountPath": "/local",
                            "name": "local-volume"
                        }
                    ]
                }
            ]

            d_job['spec']['template']['spec']['containers'].append({
                "name": "publish",
                "image": "fnndsc/pman-swift-publisher",
                "env": [
                    {
                        "name": "SWIFT_KEY",
                        "value": name
                    },
                    {
                        "name": "NUMBER_OF_WORKERS",
                        "value": number_of_workers
                    },
                    {
                        "name": "KUBECFG_PATH",
                        "value": "/tmp/.kube/config"
                    },
                    {
                        "name": "OPENSHIFTMGR_PROJECT",
                        "value": self.project
                    },
                    {
                        "name": "OUTGOING_DIR",
                        "value": outgoing_dir
                    }
                ],
                "command": [
                    "python3",
                    "put_data.py"
                ],
                "resources": {
                    "limits": {
                        "memory": "1024Mi",
                        "cpu": "2000m"
                    },
                    "requests": {
                        "memory": "150Mi",
                        "cpu": "250m"
                    }
                },
                "volumeMounts": [
                    {
                        "mountPath": "/share",
                        "name": "shared-volume"
                    },
                    {
                        "mountPath": "/etc/swift",
                        "name": "swift-credentials",
                        "readOnly": True
                    },
                    {
                        "name": "kubecfg-volume",
                        "mountPath": "/tmp/.kube/",
                        "readOnly": True
                    },
                    {
                        "mountPath": "/local",
                        "name": "local-volume"
                    }
                ]
            })
            d_job['spec']['template']['spec']['volumes'] = [
                {
                    "name": "shared-volume",
                    "persistentVolumeClaim": {
                        "claimName": name + "-storage-claim"
                    }
                },
                {
                    "name": "swift-credentials",
                    "secret": {
                        "secretName": "swift-credentials"
                    }
                },
                {
                    "name": "kubecfg-volume",
                    "secret": {
                        "secretName": "kubecfg"
                    }
                },
                {
                    "name": "local-volume",
                    "emptyDir": {}
                }
            ]
        else: # os.environ.get('STORAGE_TYPE') == 'hostPath'
            d_job['spec']['template']['spec']['volumes'] = [
                {
                    "name": "shared-volume",
                    "hostPath": {
                        "path": "/tmp/share/key-" + name
                    }
                }
            ]

        job = self.kube_v1_batch_client.create_namespaced_job(namespace=self.project, body=d_job)
        return job

    def create_pod(self, image, name, command):
        """
        Create a pod
        """
        pod_str = """
apiVersion: v1
kind: Pod
metadata:
    name: {name}
spec:
    restartPolicy: Never
    containers:
    - name: {name}
      image: {image}
      command: {command}  
""".format(name=name, image=image, command=command)
        pod_yaml = yaml.load(pod_str)
        pod = self.kube_client.create_namespaced_pod(namespace=self.project, body=pod_yaml)
        return pod

    def get_pod_status(self, name):
        """
        Get a pod's status
        """
        log = self.kube_client.read_namespaced_pod_status(namespace=self.project, name=name)
        return log

    def get_pod_log(self, name, container_name=None):
        """
        Get a pod log
        """
        if container_name:
            log = self.kube_client.read_namespaced_pod_log(namespace=self.project, name=name, container=container_name)
        else:
            log = self.kube_client.read_namespaced_pod_log(namespace=self.project, name=name)
        return log

    def get_job(self, name):
        """
        Get the previously scheduled job object
        """
        return self.kube_v1_batch_client.read_namespaced_job(name, self.project)

    def remove_job(self, name):
        """
        Remove a previously scheduled job
        """
        body = k_client.V1DeleteOptions(propagation_policy='Background')
        self.kube_v1_batch_client.delete_namespaced_job(name, body=body, namespace=self.project)

    def remove_pod(self, name):
        """
        Remove a previously scheduled pod
        """
        self.kube_client.delete_namespaced_pod(name, self.project, {})

    def state(self, name):
        """
        Return the state of a previously scheduled job
        """
        job = self.get_job(name)
        message = None
        state = None
        reason = None
        if job.status.conditions:
            for condition in job.status.conditions:
                if condition.type == 'Failed' and condition.status == 'True':
                    message = 'started'
                    reason = condition.reason
                    state = 'failed'
                    break
        if not state:
            if job.status.completion_time and job.status.succeeded > 0:
                message = 'finished'
                state = 'complete'
            elif job.status.active > 0:
                message = 'started'
                state = 'running'
            else:
                message = 'inactive'
                state = 'inactive'

        return {'Status': {'Message': message,
                                    'State': state,
                                    'Reason': reason,
                                    'Active': job.status.active,
                                    'Failed': job.status.failed,
                                    'Succeeded': job.status.succeeded,
                                    'StartTime': job.status.start_time,
                                    'CompletionTime': job.status.completion_time}}
   

    def get_job_pod_logs(self, pod_name, jid):
        """
        Returns the concatenated log of all 3 containers part of job template.
        :param str pod_name: job-id of OpenShift job.  
        :return: str log: Combined log of all the containers in pod.
        """
        # Assumption is pod is always going to have a init-storage, container with job-id, publish container, if not we just return log for job container.
        # TODO: @ravig: Think of a better way to abstract out logs in case of multiple pods running parallelly.

        job_container_log = self.get_pod_log(pod_name, jid)
        try: 
            init_container_log = self.get_pod_log(pod_name, 'init-storage')
        except ApiException:
            # We don't have init container, we assume init-storage and publish containers are not available.
            return pod_name + ":" + job_container_log
        publish_container_log = self.get_pod_log(pod_name, 'publish')
        return pod_name + ":" + "init_container log:" + init_container_log + "plugin_container log:" +\
               job_container_log + "publish_container log:" + publish_container_log

    def get_pod_names_in_job(self, job_id):
        """
        Returns names of all the pods created as part of job
        :param str job_id: job-id of OpenShift job
        :return: 
        """
        pod_names = []
        # job-name becomes selector based on which we can choose jobs that we created just now.
        pods = self.kube_client.list_namespaced_pod(self.project, label_selector='job-name='+job_id)
        for _, pod_item in enumerate(pods.items):
            pod_names.append(pod_item.metadata.name)
        return pod_names

    def create_pvc(self, job_id):
        """
        Create a Persistent Volume Claim (PVC) with the name 'jid'-storage-claim
        :param str job_id: job-id of the Openshift job
        :return:
        """
        d_pvc = {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {
                "name": job_id+"-storage-claim"
            },
            "spec": {
                "accessModes": [
                    "ReadWriteMany"
                ],
                "resources": {
                    "requests": {
                        "storage": "5Gi"
                    }
                }
            }
        }
        return self.kube_client.create_namespaced_persistent_volume_claim(self.project, body=d_pvc)
    
    def remove_pvc(self, job_id):
        """
        Remove pvc created
        """
        pvc_name = job_id+"-storage-claim"
        body = k_client.V1DeleteOptions()
        self.kube_client.delete_namespaced_persistent_volume_claim(pvc_name, self.project, body=body)



#!/usr/bin/env python3
#
# (c) 2020 Fetal-Neonatal Neuroimaging & Developmental Science Center
#                   Boston Children's Hospital
#
#              http://childrenshospital.org/FNNDSC/
#                        dev@babyMRI.org
#

import  sys
import  os
from    distutils.sysconfig     import get_python_lib

from    argparse                import RawTextHelpFormatter
from    argparse                import ArgumentParser
import  shutil

import  pfmisc
from    pfmisc._colors          import Colors
from    pfmisc.C_snode          import  *


str_version = "1.0.0.2"
str_name    = 'db_clean'
str_desc    = Colors.CYAN + """
     _ _          _
    | | |        | |
  __| | |__   ___| | ___  __ _ _ __    _ __  _   _
 / _` | '_ \ / __| |/ _ \/ _` | '_ \  | '_ \| | | |
| (_| | |_) | (__| |  __/ (_| | | | |_| |_) | |_| |
 \__,_|_.__/ \___|_|\___|\__,_|_| |_(_) .__/ \__, |
         ______                       | |     __/ |
        |______|                      |_|    |___/


    font generated by:
    http://patorjk.com/software/taag/#p=display&f=Doom&t=db_clean


                             DataBase cleaner

           A simple stand alone script that merely cleans the
           pman database.

                              -- version """ + \
           Colors.YELLOW + str_version + Colors.CYAN + """ --

    'db_clean.py' is very simple app that performs the same set
    of operations that `pman` doesn in cleaning its internal
    file-system based state DB.

    The need for the script was simply to address a quick and
    simple mechanism for effecting a `pman` data base clean after
    CUBE integration tests.


""" + Colors.NO_COLOUR

class db_clean:
    def __init__(self, **kwargs):
        """
        Simple constructor
        """
        self.str_DBpath         = "/tmp/pman"
        self.__name__           = "db_clean"

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.verbosity          = 1

        for key,val in kwargs.items():
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'verbosity':      self.verbosity      = int(val)


        self.dp                 = pfmisc.debug(
                                            verbosity   = self.verbosity,
                                            within      = self.__name__,
                                            syslog      = False
                                )
        self.dp.methodcol       = 20

    def do(self):
        """
        Main method to clear the DB.

        Essentially, this simply overwrites the file-based
        state information that `pman` maintains.

        """

        def saveToDiskAsJSON(tree_DB):
            tree_DB.tree_save(
                startPath       = '/',
                pathDiskRoot    = self.str_DBpath,
                failOnDirExist  = False,
                saveJSON        = True,
                savePickle      = False)

        self.dp.qprint('%50s%8s' % ('Clearing internal memory DB...', 
                                    '[ done ]'))
        tree_DB = C_stree()
        self.dp.qprint('%50s%8s' % ('Removing DB from persistent storage...', 
                                    '[ done ]'))
        if os.path.isdir(self.str_DBpath):
            shutil.rmtree(self.str_DBpath, ignore_errors=True)
        self.dp.qprint('%50s%8s' % ('Saving empty DB to peristent storage...',
                                    '[ done ]'))
        saveToDiskAsJSON(tree_DB)

def synopsis(ab_shortOnly = False):
    scriptName = os.path.basename(sys.argv[0])
    shortSynopsis =  '''
    NAME

	    db_clean.py

        - clean a `pman` database

    SYNOPSIS

            db_clean.py                                             \\
                [--DBpath <str_location>]                           \\
                [--synopsis]                                        \\
                [--desc]                                            \\
                [--version]                                         \\
                [--verbosity <level>]

    BRIEF EXAMPLE

            db_clean.py

    '''

    description =  '''
    DESCRIPTION

        `db_clean.py` simply cleans a `pman` filesystem state "tree".

    ARGS

        [--DBpath <str_location>]
        The location on persistent storage of the state "data base".
        Typically this is `/tmp/pman`.

        [-x|--desc]
        Provide an overview help page.

        [-y|--synopsis]
        Provide a synopsis help summary.

        [--version]
        Print internal version number and exit.

        [-v|--verbosity <level>]
        Set the verbosity level. "0" typically means no/minimal output. 


    EXAMPLES

            db_clean.py --DBpath /tmp/pman

    '''
    if ab_shortOnly:
        return shortSynopsis
    else:
        return shortSynopsis + description

parser  = ArgumentParser(description = str_desc, formatter_class = RawTextHelpFormatter)


parser.add_argument(
    "-p", "--DBpath",
    help    = "location of the DB path",
    dest    = 'DBpath',
    default = "/tmp/pman"
)
parser.add_argument(
    '--version',
    help    = 'if specified, print version number',
    dest    = 'b_version',
    action  = 'store_true',
    default = False
)
parser.add_argument(
    "-v", "--verbosity",
    help    = "verbosity level for app",
    dest    = 'verbosity',
    default = "1")
parser.add_argument(
    "-x", "--desc",
    help    = "long synopsis",
    dest    = 'desc',
    action  = 'store_true',
    default = False
)
parser.add_argument(
    "-y", "--synopsis",
    help    = "short synopsis",
    dest    = 'synopsis',
    action  = 'store_true',
    default = False
)

args = parser.parse_args()


if args.desc or args.synopsis:
    print(str_desc)
    if args.desc:
        str_help     = synopsis(False)
    if args.synopsis:
        str_help     = synopsis(True)
    print(str_help)
    sys.exit(1)

if args.b_version:
    print("Version: %s" % str_version)
    sys.exit(1)

clean   = db_clean(
            DBpath      = args.DBpath,
            verbosity   = args.verbosity
        )
clean.do()



import sys
# Make sure we are running python3.5+
if 10 * sys.version_info[0]  + sys.version_info[1] < 35:
    sys.exit("Sorry, only Python 3.5+ is supported.")

from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
      name             =   'pman',
      version          =   '2.1.0.6',
      description      =   '(Python) Process Manager',
      long_description =   readme(),
      author           =   'Rudolph Pienaar',
      author_email     =   'rudolph.pienaar@gmail.com',
      url              =   'https://github.com/FNNDSC/pman',
      packages         =   ['pman'],
      install_requires =   ['pycurl', 'pyzmq', 'webob', 'pudb', 'psutil', 'docker', 'openshift', 'pfmisc', 'ipaddress', 'fasteners', 'kubernetes', 'python-keystoneclient'],
      test_suite       =   'nose.collector',
      tests_require    =   ['nose'],
      scripts          =   ['bin/pman', 'bin/db_clean.py'],
      license          =   'MIT',
      zip_safe         =   False
     )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import  abc
import  time
import  os

import  threading
import  zmq
from    webob           import  Response
import  psutil

import  queue
from    functools       import  partial
import  platform

import  multiprocessing
import  inspect

import  json
import  ast
import  shutil
import  datetime
import  socket
import  uuid

import  pfmisc

# pman local dependencies
try:
    from    .openshiftmgr   import *
except:
    from    openshiftmgr    import *
try:
    from    .crunner        import *
except:
    from    crunner         import *

from pfmisc.Auth            import Auth
from pfmisc.C_snode         import  *
from pfmisc._colors         import  Colors

import  docker
import  pudb
import  pprint
from kubernetes.client.rest import ApiException

str_devNotes = """

    08 June 2017
    * NOTE: The zmq socket *always* sends back HTTP formatted headers around
            the response string. The listening object (usually pfurl) should
            *NOT* parse this with --httpResponseBodyParse!

    10 May 2017
    *   Should methods in the listener be functors? Certain methods, such as
        'run' and 'status' need specialized implementations based on a run
        environment. This run environment is not known by the listener when
        it starts, but can be specified at payload parsing by the process()
        method. This, a method such as

            t_run_process()

        might need at arbitrary call time to be specialized to some external
        condition set (say by running as a container). Naively, this can be
        parsed in the message and thread redirected to

            t_run_process_swarm()

        for example.

        Would a functor type approach be useful at all?

"""

class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stopper = threading.Event()

    def stopit(self):
        self._stopper.set()

    def stopped(self):
        return self._stopper.isSet()


class pman(object):
    """
    The server class for the pman (process manager) server

    """
    __metaclass__   = abc.ABCMeta

    def col2_print(self, str_left, str_right, level = 1):
        self.dp.qprint(Colors.WHITE +
              ('%*s' % (self.LC, str_left)),
              end       = '',
              level     = level,
              syslog    = False)
        self.dp.qprint(Colors.CYAN +
              ('%*s' % (self.RC, str_right)) + Colors.NO_COLOUR,
              level     = level,
              syslog    = False)

    def __init__(self, **kwargs):
        """
        Constructor
        """
        # An encapsulating object
        self.within             = None

        # Description
        self.str_desc           = ""
        self.str_name           = ""
        self.str_version        = ""
        self.__name__           = 'pman'

        # The main server function
        self.threaded_server    = None

        # The listener thread array -- each element of this array is a
        # threaded listener object
        self.l_listener         = []
        self.listenerSleep      = 0.1

        # The fileIO threaded object
        self.fileIO             = None

        # DB
        self.b_clearDB          = False
        self.str_DBpath         = '/tmp/pman'
        self.ptree             = C_stree()
        self.str_fileio         = 'json'
        self.DBsavePeriod       = 60

        # Comms
        self.str_protocol       = "tcp"
        self.str_IP             = "127.0.0.1"
        self.str_port           = "5010"
        self.router_raw         = 0
        self.listeners          = 1
        self.b_http             = False
        self.socket_front       = None
        self.socket_back        = None

        # Job info
        self.auid               = ''
        self.jid                = ''
        self.container_env      = ''

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.pp                 = pprint.PrettyPrinter(indent=4)
        self.verbosity          = 1

        # Authentication parameters
        self.b_tokenAuth        = False
        self.authModule         = None

        for key,val in kwargs.items():
            if key == 'protocol':       self.str_protocol   = val
            if key == 'IP':             self.str_IP         = val
            if key == 'port':           self.str_port       = val
            if key == 'raw':            self.router_raw     = int(val)
            if key == 'listeners':      self.listeners      = int(val)
            if key == 'listenerSleep':  self.listenerSleep  = float(val)
            if key == 'DBsavePeriod':   self.DBsavePeriod   = int(val)
            if key == 'http':           self.b_http         = int(val)
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'clearDB':        self.b_clearDB      = val
            if key == 'desc':           self.str_desc       = val
            if key == 'name':           self.str_name       = val
            if key == 'version':        self.str_version    = val
            if key == 'containerEnv':   self.container_env  = val.lower()
            if key == 'verbosity':      self.verbosity      = int(val)
            if key == 'b_tokenAuth':    self.b_tokenAuth    = val
            if key == 'str_tokenPath':
                if self.b_tokenAuth:
                    self.authModule = Auth('socket', val)
        # Screen formatting
        self.LC                 = 30
        self.RC                 = 40
        self.dp                 = pfmisc.debug(
                                            verbosity   = self.verbosity,
                                            debugFile   = self.str_debugFile,
                                            debugToFile = self.b_debugToFile,
                                            within      = self.__name__)

        if self.b_clearDB and os.path.isdir(self.str_DBpath):
            shutil.rmtree(self.str_DBpath)

        self.dp.qprint(self.str_desc, level = 1)

        self.col2_print('Server is listening on',
                        '%s://%s:%s' % (self.str_protocol, self.str_IP, self.str_port))
        self.col2_print('Router raw mode',                  str(self.router_raw))
        self.col2_print('HTTP response back mode',          str(self.b_http))
        self.col2_print('listener sleep',                   str(self.listenerSleep))

        # Create the main internal DB data structure/abstraction
        self.ptree             = C_stree()

        # Read the DB from HDD
        self.DB_fileIO(cmd = 'load')

        # Setup zmq context
        self.zmq_context        = zmq.Context()

    def DB_read(self, **kwargs):
        """
        Read the DB from filesystem. If DB does not exist on filesystem,
        create an empty DB and save to filesystem.
        """
        if os.path.isdir(self.str_DBpath):
            self.dp.qprint("Reading pman DB from disk...\n")
            self.ptree = C_stree.tree_load(
                pathDiskRoot    = self.str_DBpath,
                loadJSON        = True,
                loadPickle      = False)
            self.dp.qprint("pman DB read from disk...\n")
            self.col2_print('Reading pman DB from disk:', 'OK')
        else:
            P = self.ptree
            # P.cd('/')
            # P.mkdir('proc')
            P.tree_save(
                startPath       = '/',
                pathDiskRoot    = self.str_DBpath,
                failOnDirExist  = False,
                saveJSON        = True,
                savePickle      = False
            )
            self.col2_print('Reading pman DB from disk:',
                            'No DB found... creating empty default DB')
        self.dp.qprint(Colors.NO_COLOUR, end='')

    def DB_fileIO(self, **kwargs):
        """
        Process DB file IO requests. Typically these control the
        DB -- save or load.
        """
        str_cmd     = 'save'
        str_DBpath  = self.str_DBpath
        tree_DB     = self.ptree

        def loadFromDiskAsJSON():
            tree_DB = C_stree.tree_load(
                startPath       = '/',
                pathDiskRoot    = str_DBpath,
                failOnDirExist  = False,
                loadJSON        = True,
                loadPickle      = False)
            return tree_DB

        def loadFromDiskAsPickle():
            tree_DB = C_stree.tree_load(
                startPath       = '/',
                pathDiskRoot    = str_DBpath,
                failOnDirExist  = False,
                loadJSON        = False,
                loadPickle      = True)
            return tree_DB

        def saveToDiskAsJSON(tree_DB):
            tree_DB.tree_save(
                startPath       = '/',
                pathDiskRoot    = str_DBpath,
                failOnDirExist  = False,
                saveJSON        = True,
                savePickle      = False)

        def saveToDiskAsPickle(tree_DB):
            tree_DB.tree_save(
                startPath       = '/',
                pathDiskRoot    = str_DBpath,
                failOnDirExist  = False,
                saveJSON        = False,
                savePickle      = True)

        for k,v in kwargs.items():
            if k == 'cmd':      str_cmd             = v
            if k == 'fileio':   self.str_fileio     = v
            if k == 'dbpath':   str_DBpath          = v
            if k == 'db':       tree_DB             = v

        # self.dp.qprint('cmd      = %s' % str_cmd)
        # self.dp.qprint('fileio   = %s' % self.str_fileio)
        # self.dp.qprint('dbpath   = %s' % str_DBpath)

        if str_cmd == 'clear':
            # This wipes the existing DB both in memory
            # and in disk storage.
            self.dp.qprint('Clearing internal memory DB...')
            tree_DB = C_stree()
            self.dp.qprint('Removing DB from persistent storage...')
            if os.path.isdir(str_DBpath):
                shutil.rmtree(str_DBpath, ignore_errors=True)
            self.dp.qprint('Saving empty DB to peristent storage')
            saveToDiskAsJSON(tree_DB)

        if str_cmd == 'save':
            if os.path.isdir(str_DBpath):
                shutil.rmtree(str_DBpath, ignore_errors=True)
            #print(tree_DB)
            if self.str_fileio   == 'json':     saveToDiskAsJSON(tree_DB)
            if self.str_fileio   == 'pickle':   saveToDiskAsPickle(tree_DB)

        if str_cmd == 'load':
            if os.path.isdir(str_DBpath):
                self.dp.qprint("Reading pman DB from disk...\n")
                if self.str_fileio   == 'json':     tree_DB = loadFromDiskAsJSON()
                if self.str_fileio   == 'pickle':   tree_DB = loadFromDiskAsPickle()
                self.dp.qprint("Pre-existing DB found at %s..." % str_DBpath)
                self.ptree         = tree_DB
                self.ptree.cd('/')
                self.dp.qprint('DB root nodes:\n%s' % self.ptree.str_lsnode())
            else:
                saveToDiskAsJSON(tree_DB)
                self.col2_print('Reading pman DB from disk:',
                                'No DB found... creating empty default DB')
            self.dp.qprint(Colors.NO_COLOUR, end='')
        self.ptree  = tree_DB

    def thread_serve(self):
        """
        Serve the 'start' method in a thread.
        :return:
        """
        self.threaded_server = StoppableThread(target=self.start)
        self.threaded_server.start()

        while not self.threaded_server.stopped():
            time.sleep(1)

        # Stop the listeners...
        self.dp.qprint("setting b_stopThread on all listeners...")
        for i in range(0, self.listeners):
            self.dp.qprint("b_stopThread on listener %d and executing join()..." % i)
            self.l_listener[i].b_stopThread = True
            self.l_listener[i].join()

        # Stop the fileIO
        self.fileIO.b_stopThread    = True
        self.dp.qprint("b_stopThread on fileIO executing join()...")
        self.fileIO.join()

        self.dp.qprint("Shutting down the zmq infrastructure...")
        try:
            self.dp.qprint('calling self.socket_back.close()')
            self.socket_back.close()
        except:
            self.dp.qprint('Caught exception in closing back socket')

        try:
            self.dp.qprint('calling self.socket_front.close()')
            self.socket_front.close()
        except zmq.error.ZMQError:
            self.dp.qprint('Caught exception in closing front socket...')

        self.dp.qprint('calling zmq_context.term()')
        # self.zmq_context.term()

        self.dp.qprint("calling join() on all this thread...")
        self.threaded_server.join()
        self.dp.qprint("shutdown successful...")

    def start(self):
        """
            Main execution.

            * Instantiate several 'listener' worker threads
                **  'listener' threads are used to process input from external
                    processes. In turn, 'listener' threads can thread out
                    'crunner' threads that actually "run" the job.
            * Instantiate a job poller thread
                **  'poller' examines the internal DB entries and regularly
                    queries the system process table, tracking if jobs
                    are still running.
        """
        self.dp.qprint('Starting %d Listener threads' % self.listeners)

        # Front facing socket to accept client connections.
        self.socket_front = self.zmq_context.socket(zmq.ROUTER)
        self.socket_front.router_raw = self.router_raw
        self.socket_front.setsockopt(zmq.LINGER, 1)
        self.socket_front.bind('%s://%s:%s' % (self.str_protocol,
                                          self.str_IP,
                                          self.str_port)
                          )

        # Backend socket to distribute work.
        self.socket_back = self.zmq_context.socket(zmq.DEALER)
        self.socket_back.setsockopt(zmq.LINGER, 1)
        self.socket_back.bind('inproc://backend')

        # Start the 'fileIO' thread
        self.fileIO      = FileIO(      DB          = self.ptree,
                                        timeout     = self.DBsavePeriod,
                                        within      = self,
                                        debugFile   = self.str_debugFile,
                                        verbosity   = self.verbosity,
                                        debugToFile = self.b_debugToFile)
        self.fileIO.start()


        # Start the 'listener' workers... keep track of each
        # listener instance so that we can selectively stop
        # them later.
        for i in range(0, self.listeners):
            self.l_listener.append(Listener(
                                    id              = i,
                                    context         = self.zmq_context,
                                    DB              = self.ptree,
                                    DBpath          = self.str_DBpath,
                                    http            = self.b_http,
                                    containerEnv    = self.container_env,
                                    within          = self,
                                    listenerSleep   = self.listenerSleep,
                                    verbosity       = self.verbosity,
                                    debugToFile     = self.b_debugToFile,
                                    debugFile       = self.str_debugFile,
                                    b_tokenAuth     = self.b_tokenAuth,
                                    authModule      = self.authModule))
            self.l_listener[i].start()

        # Use built in queue device to distribute requests among workers.
        # What queue device does internally is,
        #   1. Read a client's socket ID and request.
        #   2. Send socket ID and request to a worker.
        #   3. Read a client's socket ID and result from a worker.
        #   4. Route result back to the client using socket ID.
        self.dp.qprint("*******before  zmq.device!!!")
        try:
            zmq.device(zmq.QUEUE, self.socket_front, self.socket_back)
        except:
            self.dp.qprint('Hmmm... some error was caught on shutting down the zmq.device...')
        self.dp.qprint("*******after zmq.device!!!")

    def __iter__(self):
        yield('Feed', dict(self.ptree.snode_root))

    # @abc.abstractmethod
    # def create(self, **kwargs):
    #     """Create a new tree
    #
    #     """

    def __str__(self):
        """Print
        """
        return str(self.ptree.snode_root)

    @property
    def stree(self):
        """STree Getter"""
        return self.ptree

    @stree.setter
    def stree(self, value):
        """STree Getter"""
        self.ptree = value

class FileIO(threading.Thread):
    """
    A class that periodically saves the database from memory out to disk.
    """

    def __init__(self, **kwargs):
        self.__name             = "FileIO"
        self.b_http             = False

        self.str_DBpath         = "/tmp/pman"

        self.timeout            = 60
        self.within             = None

        self.b_stopThread       = False
        self.verbosity          = 1

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.pp                 = pprint.PrettyPrinter(indent=4)

        for key,val in kwargs.items():
            if key == 'DB':             self.ptree          = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'timeout':        self.timeout        = val
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val
            if key == 'verbosity':      self.verbosity      = int(val)

        self.dp                 = pfmisc.debug(
                                        verbosity   = self.verbosity,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile,
                                        within      = self.__name)

        threading.Thread.__init__(self)

    def run(self):
        """ Main execution. """
        # Socket to communicate with front facing server.
        while not self.b_stopThread:
            # self.dp.qprint('Saving DB as type "%s" to "%s"...' % (
            #     self.within.str_fileio,
            #     self.within.str_DBpath
            # ))
            self.within.DB_fileIO(cmd = 'save')
            # self.dp.qprint('DB saved...')
            for second in range(0, self.timeout):
                if not self.b_stopThread:
                    time.sleep(1)
                else:
                    break

        self.dp.qprint('returning from FileIO run method...')
        # raise ValueError('FileIO thread terminated.')

class Listener(threading.Thread):
    """ Listeners accept communication requests from front facing server.
        Parse input text streams and act accordingly. """

    def __init__(self, **kwargs):
        self.__name             = "Listener"
        self.b_http             = False

        self.poller             = None
        self.str_DBpath         = "/tmp/pman"
        self.str_jobRootDir     = ''

        self.listenerSleep      = 0.1
        self.verbosity          = 1

        self.jid                = ''
        self.auid               = ''

        self.within             = None
        self.b_stopThread       = False

        self.openshiftmgr       = None

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.pp                 = pprint.PrettyPrinter(indent=4)

        for key,val in kwargs.items():
            if key == 'context':        self.zmq_context    = val
            if key == 'listenerSleep':  self.listenerSleep  = float(val)
            if key == 'id':             self.worker_id      = val
            # if key == 'DB':             self.ptree          = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'http':           self.b_http         = val
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val
            if key == 'containerEnv':   self.container_env  = val
            if key == 'verbosity':      self.verbosity      = int(val)
            if key == 'b_tokenAuth':    self.b_tokenAuth    = val
            if key == 'authModule':     self.authModule     = val

        self.dp                 = pfmisc.debug(
                                        verbosity   = self.verbosity,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile,
                                        within      = self.__name)

        threading.Thread.__init__(self)
        # logging.debug('leaving __init__')

    def df_print(self, adict):
        """
        Return a nicely formatted string representation of a dictionary
        """
        return self.pp.pformat(adict).strip()

    def run(self):
        """ Main execution. """
        # Socket to communicate with front facing server.
        self.dp.qprint('starting...')
        socket = self.zmq_context.socket(zmq.DEALER)
        socket.connect('inproc://backend')

        b_requestWaiting        = False
        resultFromProcessing    = False
        request                 = ""
        client_id               = -1
        self.dp.qprint(Colors.BROWN + "Listener ID - %s: run() - Ready to serve..." % self.worker_id, level = 1)
        while not self.b_stopThread:

            # wait (non blocking) for input on socket
            try:
                client_id, request  = socket.recv_multipart(flags = zmq.NOBLOCK)
                self.dp.qprint('Received %s from client_id: %s' % (request, client_id))
                b_requestWaiting    = True
            except zmq.Again as e:
                if self.listenerSleep:
                    time.sleep(0.1)
                else:
                    pass

            if b_requestWaiting:
                self.dp.qprint(Colors.BROWN + 'Listener ID - %s: run() - Received comms from client.' % (self.worker_id))
                self.dp.qprint(Colors.BROWN + 'Client sends: %s' % (request))

                resultFromProcessing    = self.process(request)
                if resultFromProcessing:
                    self.dp.qprint(Colors.BROWN + 'Listener ID - %s: run() - Sending response to client.' %
                                   (self.worker_id))
                    self.dp.qprint('JSON formatted response:')
                    str_payload = json.dumps(resultFromProcessing, sort_keys=False, indent=4)
                    self.dp.qprint(Colors.LIGHT_CYAN + str_payload)
                    self.dp.qprint(Colors.BROWN + 'len = %d chars' % len(str_payload))
                    socket.send(client_id, zmq.SNDMORE)
                    if self.b_http:
                        str_contentType     = "application/html"
                        res                 = Response(str_payload)
                        res.content_type    = str_contentType

                        str_HTTPpre         = "HTTP/1.1 "
                        str_res             = "%s%s" % (str_HTTPpre, str(res))
                        str_res             = str_res.replace("UTF-8", "UTF-8\nAccess-Control-Allow-Origin: *")
                        self.dp.qprint('HTML response')
                        self.dp.qprint(str_res.encode())
                        socket.send(str_res.encode())
                    else:
                        str_contentType     = "application/json"
                        res                 = Response(str_payload)
                        res.content_type    = str_contentType
                        str_HTTPpre         = "HTTP/1.1 "
                        str_res             = '%s%s' % (str_HTTPpre, (res))
                        self.dp.qprint(str_res)
                        socket.send_string(str_res)
            b_requestWaiting    = False
        self.dp.qprint('Listener ID - %s: Returning from run()...' % self.worker_id)
        # raise('Listener ID - %s: Thread terminated' % self.worker_id)
        return True

    def t_search_process(self, *args, **kwargs):
        """

        Search

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In search process...")

        d_request   = {}
        d_ret       = {}
        hits        = 0

        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta          = d_request['meta']

        b_pathSpec      = False
        str_path        = ""
        if 'path' in d_meta:
            b_pathSpec  = True
            str_path    = d_meta['path']

        b_jobSpec       = False
        str_jobSpec    = ""
        if 'job' in d_meta:
            b_jobSpec   = True
            str_jobSpec = d_meta['job']

        b_fieldSpec    = False
        str_fieldSpec  = ""
        if 'field' in d_meta:
            b_fieldSpec = True
            str_fieldSpec = d_meta['field']

        b_whenSpec      = False
        str_whenSpec    = "end"
        if 'when' in d_meta:
            b_whenSpec = True
            str_whenSpec = d_meta['when']

        self.dp.qprint(d_meta)
        self.dp.qprint(b_pathSpec)
        str_fileName    = d_meta['key']
        str_target      = d_meta['value']
        p               = self.within.ptree
        str_origDir     = p.cwd()
        str_pathOrig    = str_path
        for r in self.within.ptree.lstr_lsnode('/'):
            if p.cd('/' + r)['status']:
                str_val = p.cat(str_fileName)
                if str_val == str_target:
                    if not b_pathSpec:
                        str_path            = '/api/v1/' + r + '/' + str_fileName
                    else:
                        str_path            = '/api/v1/' + r + str_pathOrig
                        if str_path[-1] == '/': str_path = str_path[:-1]
                    if b_jobSpec:
                        str_path            = '/api/v1/' + r +              '/' + \
                                                str_whenSpec +              '/' + \
                                                str_jobSpec +               '/' + \
                                                '%sInfo' % str_whenSpec +   '/' + \
                                                str_jobSpec +               '/' + \
                                                str_fieldSpec
                    d_ret[str(hits)]    = {}
                    d_ret[str(hits)]    = self.DB_get(path = str_path)
                    hits               += 1
        p.cd(str_origDir)

        return {"d_ret":    d_ret,
                "status":   bool(hits)}

    def t_info_process(self, *args, **kwargs):
        """

        Check if the job corresponding to the search pattern is "done".

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In info process...")

        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_search    = self.t_search_process(request = d_request)['d_ret']

        p = self.within.ptree
        for j in d_search.keys():
            d_j = d_search[j]
            for job in d_j.keys():
                str_pathStart       = '/api/v1/' + job + '/startInfo'
                str_pathEnd         = '/api/v1/' + job + '/endInfo'
                d_ret[str(hits)+'.0']    = {}
                d_ret[str(hits)+'.0']    = self.DB_get(path = str_pathStart)
                d_ret[str(hits)+'.1']    = {}
                d_ret[str(hits)+'.1']    = self.DB_get(path = str_pathEnd)
                hits               += 1
        if not hits:
            d_ret                   = {
                "-1":   {
                    "noJobFound":   {
                        "endInfo":  {"allJobsDone": None}
                    }
                }
            }
        else:
            b_status            = True
        return {"d_ret":    d_ret,
                "status":   b_status}

    def t_quit_process(self, *args, **kwargs):
        """
        Process the 'quit' POST directive. This might appear counter-intuitive
        at first glance since the 'get' is the result of a REST POST, but is
        logically consistent within the semantics of this system.
        """
        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v
        d_meta      = d_request['meta']
        if 'saveDB' in d_meta.keys():
            self.dp.qprint("Saving DB...")
            self.within.DB_fileIO(cmd = 'save')

        self.dp.qprint('calling threaded_server.stop()')
        self.within.threaded_server.stopit()
        self.dp.qprint('called threaded_server.stop()')

        return {'d_ret':    d_ret,
                'status':   True}

    def t_get_process(self, *args, **kwargs):
        """
        Process the 'get' POST directive. This might appear counter-intuitive
        at first glance since the 'get' is the result of a REST POST, but is
        logically consistent within the semantics of this system.
        """
        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v
        d_meta      = d_request['meta']
        str_path    = '/api/v1' + d_meta['path']
        d_ret       = self.DB_get(path  = str_path)
        return {'d_ret':    d_ret,
                'status':   True}

    def t_DBctl_process(self, *args, **kwargs):
        """
        Entry point for internal DB control processing.
        """
        tree_DB     = self.within.ptree
        d_request   = {}
        d_ret       = {}
        b_status    = False
        str_fileio  = self.within.str_fileio
        str_DBpath  = self.within.str_DBpath

        for k, v in kwargs.items():
            if k == 'request':      d_request   = v
        d_meta      = d_request['meta']

        if 'do'         in d_meta:  str_do          = d_meta['do']
        if 'dbpath'     in d_meta:  str_DBpath      = d_meta['dbpath']
        if 'fileio'     in d_meta:  str_fileio      = d_meta['fileio']

        self.within.DB_fileIO(
                        cmd         = str_do,
                        fileio      = str_fileio,
                        dbpath      = str_DBpath,
                        db          = tree_DB
                        )

        # str_path    = '/api/v1' + str_DBpath
        d_ret       = self.DB_get(path  = str_DBpath)
        return {'d_ret':    d_ret,
                'status':   True}


    def t_fileiosetup_process(self, *args, **kwargs):
        """
        Setup a thread with a socket listener. Return listener address to client
        """
        self.dp.qprint("In fileiosetup process...")

        d_ret               = {}
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta  = d_request['meta']

        d_ret['fileioIP']   = "%s" % self.within.str_IP
        d_ret['fileioport'] = "%s" % (int(self.within.str_port) + self.worker_id)
        d_ret['serveforever']=d_meta['serveforever']

        d_args              = {
                                'ip':    d_ret['fileioIP'],
                                'port':  d_ret['fileioport']
                               }

        server              = ThreadedHTTPServer((d_args['ip'], int(d_args['port'])), StoreHandler)
        server.setup(args   = d_args)
        self.dp.qprint("serveforever = %d" % d_meta['serveforever'])
        b_serveforever      = False
        if 'serveforever' in d_meta.keys():
            b_serveforever  = d_meta['serveforever']

        if b_serveforever:
            self.dp.qprint("about to serve_forever()...")
            server.serve_forever()
        else:
            self.dp.qprint("about to handle_request()...")
            server.handle_request()

        return {"d_ret":    d_ret,
                "status":   True}

    def job_state(self, *args, **kwargs):
        """

        Return a structure that can be further processed to determine the job's state.

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In job_state()...")

        d_request   = {}
        d_ret       = {}
        b_status    = False
        b_container = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_search    = self.t_search_process(request = d_request)['d_ret']

        p   = self.within.ptree
        Ts  = C_stree()
        Te  = C_stree()
        for j in d_search.keys():
            d_j = d_search[j]
            for job in d_j.keys():
                str_pathStart       = '/api/v1/' + job + '/start'
                str_pathEnd         = '/api/v1/' + job + '/end'

                d_start             = self.DB_get(path = str_pathStart)
                d_end               = self.DB_get(path = str_pathEnd)
                Ts.initFromDict(d_start)
                Te.initFromDict(d_end)

                self.dp.qprint("Ts.cwd = %s " % Ts.cwd())
                self.dp.qprint(Ts)
                self.dp.qprint("Te.cwd = %s " % Te.cwd())
                self.dp.qprint(Te)

                l_subJobsStart      = []
                if Ts.cd('/%s/start' % job)['status']:
                    l_subJobsStart  = Ts.lstr_lsnode()
                    l_subJobsStart  = list(map(int, l_subJobsStart))
                    l_subJobsStart.sort()
                    self.dp.qprint("l_subJobsStart  (pre) = %s" % l_subJobsStart)
                    if len(l_subJobsStart) > 1: l_subJobsStart  = l_subJobsStart[:-1]

                l_subJobsEnd        = []
                if Te.cd('/%s/end' % job)['status']:
                    l_subJobsEnd    = Te.lstr_lsnode()
                    l_subJobsEnd    = list(map(int, l_subJobsEnd))
                    l_subJobsEnd.sort()
                    self.dp.qprint("l_subJobsEnd    (pre) = %s " % l_subJobsEnd)
                    if len(l_subJobsEnd) > 1: l_subJobsEnd    = l_subJobsEnd[:-1]

                self.dp.qprint("l_subJobsStart (post) = %s" % l_subJobsStart)
                self.dp.qprint("l_subJobsEnd   (post) = %s" % l_subJobsEnd)

                for j in l_subJobsStart:
                    l_subJobsStart[j]   = Ts.cat('/%s/start/%d/startInfo/%d/startTrigger' % \
                                                 (job, j, j))

                # jobsEnd behaviour can be slightly different to the jobStart, particularly if
                # the job being executed is killed -- sometimes recording the "death" event of
                # the job does not happen and the job indexing ends up missing several epochs:
                #
                #           l_subJobsStart  (pre) = [0, 1, 2, 3, 4]
                #           l_subJobsEnd    (pre) = [0, 1, 3, 4]
                #
                # to assure correct returncode lookup, we always parse the latest job epoch.

                latestJob       = 0
                if len(l_subJobsEnd):
                    latestJob   = l_subJobsEnd[-1]
                    for j in list(range(0, latestJob+1)):
                        l_subJobsEnd[j]         = Te.cat('/%s/end/%s/endInfo/%d/returncode' % (job, latestJob, j))
                T_container     = False

                if p.exists('container', path = '/%s' % job):
                    T_container = C_stree()
                    p.copy(startPath = '/%s/container' % (job), destination = T_container)
                    d_ret[str(hits)+'.container']   = {"jobRoot": job, "tree": dict(T_container.snode_root)}
                else:
                    d_ret[str(hits)+'.container']   = {"jobRoot": job, "tree":      None}
                d_ret[str(hits)+'.start']       = {"jobRoot": job, "startTrigger":  l_subJobsStart}
                d_ret[str(hits)+'.end']         = {"jobRoot": job, "returncode":    l_subJobsEnd}
                hits               += 1
        if not hits:
            d_ret['-1.start']       = {"jobRoot":   None, "startTrigger":   None}
            d_ret['-1.end']         = {"jobRoot":   None, "returncode":     None}
            d_ret['-1.container']   = {"jobRoot":   None, "tree":           None}
        else:
            b_status            = True
        return {"hits":         hits,
                "d_ret":        d_ret,
                "status":       b_status}

    def t_done_process(self, *args, **kwargs):
        """

        Check if the job corresponding to the search pattern is "done".
        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In done process...")

        return self.job_state(*args, **kwargs)

    def t_status_process(self, *args, **kwargs):
        """

        This method is the main (threaded) entry point for returning
        information on the status of jobs (both active and historical)
        that have been (or are currently) managed by pman.

        Originally, the concept of "job" only extended to a command
        line process spawned off on the underlying shell. With time,
        however, this concept expanded to encompass processes that
        are containerized.

        While most (if not all) of the use of pman currently is to
        handle containerized compute, the status determination logic
        still retains the ability to query simple spawned jobs.

        The determination about whether or not a job has been
        containerized is quite simple -- a token in the internal
        job "state" memory structure (the main pman stree "DB")
        is checked -- this initial chunk of data is returned by a
        call to self.job_state() which delivers a dictionary
        representation of the jobRoot in the DB tree.

        :param args:
        :param kwargs:
        :return: dictionary of components defining job state.
        """
        self.dp.qprint("------- In status process ------------")
        status = logs = currentState = ''
        if self.container_env == 'openshift':
            self.dp.qprint('------- Processing openshift status request -----------')
            try:
                d_containerStatus       =   self.t_status_process_openshift(*args, **kwargs)
                status                  =   d_containerStatus['status']
                logs                    =   d_containerStatus['logs']
                currentState            =   d_containerStatus['currentState']
            except Exception as e:
                if isinstance(e, ApiException) and e.reason == 'Not Found':
                    status = logs = currentState = e.reason
                else:
                    raise e

            d_ret = {
                'description':   str(status),
                'l_logs':     str(logs),
                'l_status': currentState
            }
            return {
                    "d_ret":    d_ret,
                    "status":   str(currentState)
            }

        else:

            d_state     = self.job_state(*args, **kwargs)
            # {
            #     "hits":         hits,
            #     "d_ret":
            #         [<index>+'.container']   = {
            #             "jobRoot": job, "tree": dict(T_container.snode_root)
            #         },
            #     "status":       b_status
            # }

            d_ret       = d_state['d_ret']
            b_status    = d_state['status']

            d_keys      = d_ret.items()
            l_status    = []
            l_logs      = []

            #
            # The d_ret keys consist of groups of
            #
            #       *.start
            #       *.end
            #       *.container
            #
            # thus the loop grouping is number of items / 3
            #
            if '0.start' in d_ret:
                for i in range(0, int(len(d_keys)/3)):
                    try:
                        b_startEvent    = d_ret['%s.start'  % str(i)]['startTrigger'][0]
                    except:
                        b_startEvent    = False
                    try:
                        endcode     = d_ret['%s.end'    % str(i)]['returncode'][0]
                    except:
                        endcode     = None

                    # Was this a containerized job?
                    found_container = False
                    container_path = '%s.%s' % (str(i), 'container')
                    if container_path in d_state['d_ret']           and \
                        d_state['d_ret'][container_path]['tree']    and \
                        b_startEvent:

                        kwargs['d_state']   = d_state
                        kwargs['hitIndex']  = str(i)

                        str_methodSuffix = None
                        if self.container_env == 'swarm':
                            # append suffix _container to redirect to container function
                            str_methodSuffix    = 'container'
                        d_containerStatus       = eval("self.t_status_process_%s(*args, **kwargs)" % str_methodSuffix)
                        # d_ret {
                        #     'status':         d_ret['status'],              # bool
                        #     'logs':           str_logs,                     # logs from app in container
                        #     'currentState':   d_ret['d_process']['state']   # string of 'finishedSuccessfully' etc
                        # }

                        l_status.append(d_containerStatus['currentState'])
                        l_logs.append(d_containerStatus['logs'])
                        found_container = True

                    # The case for non-containerized jobs
                    if not found_container:
                        if endcode is None and not b_startEvent:
                            l_status.append('notstarted')
                        if endcode is None and b_startEvent:
                            l_status.append('started')
                        if not endcode and b_startEvent and type(endcode) is int:
                            l_status.append('finishedSuccessfully')
                        if endcode and b_startEvent:
                            l_status.append('finishedWithError')

                    self.dp.qprint('b_startEvent = %d' % b_startEvent)
                    self.dp.qprint(endcode)
                    self.dp.qprint('l_status = %s' % l_status)

        d_ret['l_status']   = l_status
        d_ret['l_logs']     = l_logs

        return {
                "d_ret":    d_ret,
                "status":   b_status
                }

    def DB_store(self, data, str_path, str_file):
        """
        In the DB memory tree, simply stores <data> to a location called
        <str_path> and a file called <str_file>.

        Explicitly separating <str_path> and <str_file> is just for
        expedience in checking up on path validity in the DB memory tree.

        This method also triggers a DB save event.

        """
        if not self.within.ptree.exists(str_file, path = str_path):
            self.within.ptree.touch('%s/%s' % (str_path, str_file), data)
            # Save DB state...
            self.within.DB_fileIO(cmd = 'save')

    def t_status_process_container_stateObject(self, *args, **kwargs):
        """
        This method processes the swarm manager state object and, if
        necessary, shuts down the service from the swarm scheduler.

        PRECONDITIONS:
        o   This method should only ever be called by t_status_process_container().

        POSTCONDITIONS:
        o   A string denoting the current state is returned.
        o   If state is complete and service still running, save state object to
            tree and remove service.
        o   Store the state object and logs in the internal DB tree!

        """

        def service_exists(str_serviceName):
            """
            Returns a bool:
                - True:     <str_serviceName> does exist
                - False:    <str_serviceName> does not exist
            """
            b_exists        = False
            client          = docker.from_env()
            try:
                service     = client.services.get(str_serviceName)
                b_exists    = True
            except:
                b_exists    = False
            return b_exists

        def service_shutDown_check():
            """
            Verifies that a docker service can be shutdown.

            Should multiple jobs have been scheduled temporally serially
            with the same jid/serviceName, then the actual service can
            only be shut down once all identical jobs have had their
            state stored.

            Returns bool:
                - True:     can shut down
                - False:    cannot shut down
            """
            ret = False
            if int(str_hitIndex) < int(d_jobState['hits'])-1:
                ret     = False
            else:
                ret      = True
            return ret

        def service_shutDown(d_serviceInfo):
            """
            Shut down a service
            """
            client          = docker.from_env()
            str_cmdShutDown = '%s --remove %s' % \
                (d_serviceInfo['managerApp'], d_serviceInfo['serviceName'])
            byte_str        = client.containers.run(
                                    '%s' % d_serviceInfo['managerImage'],
                                    str_cmdShutDown,
                                    volumes = {
                                                '/var/run/docker.sock':
                                                        {
                                                            'bind': '/var/run/docker.sock',
                                                            'mode': 'rw'
                                                        }
                                                },
                                    remove=True)
            return byte_str

        d_serviceState      = None
        d_jobState          = None
        str_hitIndex        = "0"
        str_logs            = ""
        for k,v in kwargs.items():
            if k == 'jobState':         d_jobState      = v
            if k == 'serviceState':     d_serviceState  = v
            if k == 'hitIndex':         str_hitIndex    = v
            if k == 'logs':             str_logs        = v

        if d_serviceState:

            d_ret    = self.t_status_process_state(**kwargs)
            # d_ret {
            #             'currentState':   str_currentState,
            #             'removeJob':      b_removeJob,
            #             'status':         True
            #         }

            if d_ret['removeJob']:
                str_jobRoot = d_jobState['d_ret']['%s.container' % (str_hitIndex)]['jobRoot']
                self.within.ptree.cd('/%s/container' % str_jobRoot)
                d_serviceInfo       = {
                                        'serviceName':  self.within.ptree.cat('manager/env/serviceName'),
                                        'managerImage': self.within.ptree.cat('manager/image'),
                                        'managerApp':   self.within.ptree.cat('manager/app')
                                    }
                if service_exists(d_serviceInfo['serviceName']):
                    service_shutDown(d_serviceInfo)

        return {
            'status':       True,
            'd_process':    d_ret
            }

    def t_status_process_container(self, *args, **kwargs):
        """
        Execution should only reach this method for "container"ized jobs
        status determination!

        The 'd_state' contains a dictionary representation of the container
        DB tree.

        PRECONDITIONS:
        o   Only call this method if a container structure exists
            in the relevant job tree!

        POSTCONDITIONS:
        o   If the job is completed, then shutdown the container cluster
            service.
        o   The memory container tree contains a dictionary called 'state'
            that is the state returned by the container service, as well as
            a file called 'logs' that is the stdout/stderr generated by the
            job as it ran in the container.
        """
        d_state         = None
        str_jobRoot     = ''
        str_hitIndex    = "0"
        str_logs        = ''

        for k,v in kwargs.items():
            if k == 'd_state':  d_state         = v
            if k == 'hitIndex': str_hitIndex    = v

        self.dp.qprint('checking on status using container...')
        str_jobRoot         = d_state['d_ret']['%s.container' % str_hitIndex]['jobRoot']
        self.within.ptree.cd('/%s/container' % str_jobRoot)
        str_serviceName     = self.within.ptree.cat('manager/env/serviceName')
        str_managerImage    = self.within.ptree.cat('manager/image')
        str_managerApp      = self.within.ptree.cat('manager/app')

        # Check if the state of the container service has been recorded to the data tree
        if self.within.ptree.exists('state', path = '/%s/container' % str_jobRoot):
            # If this exists, then the job has actually completed and
            # its state has been recorded in the data tree. We can simply 'cat'
            # the state from this memory dictionary
            d_serviceState  = self.within.ptree.cat('/%s/container/state' % str_jobRoot)
            if self.within.ptree.exists('logs', path = '/%s/container' % str_jobRoot):
                # The job has actually completed and its logs are recorded in the data tree
                str_logs     = self.within.ptree.cat('/%s/container/logs' % str_jobRoot)
        else:
            # Here, the manager has not been queried yet about the state of
            # the service. We need to ask the container service for this
            # state, and then record the state (and logs) in the memory
            # tree, and then "shut down" the service.
            client = docker.from_env()

            # Get the state of the service...
            str_cmdManager  = '%s --state %s' % \
                              (str_managerApp, str_serviceName)
            byte_str        = client.containers.run(
                    '%s' % str_managerImage,
                    str_cmdManager,
                    volumes =   {
                                '/var/run/docker.sock':
                                    {
                                        'bind': '/var/run/docker.sock',
                                        'mode': 'rw'
                                    }
                                },
                                remove  = True)
            d_serviceState  = json.loads(byte_str.decode())
            # Now, parse for the logs of the actual container run by the service:
            # NB: This has only really tested/used on swarm!!
            b_containerIDFound = True
            try:
                str_contID  = d_serviceState['Status']['ContainerStatus']['ContainerID']
                b_containerIDFound  = True
            except:
                b_containerIDFound  = False
            if b_containerIDFound:
                container   = client.containers.get(str_contID)
                str_logs    = container.logs()
                str_logs    = str_logs.decode()

        d_ret = self.t_status_process_container_stateObject(
                                    hitIndex        = str_hitIndex,
                                    jobState        = d_state,
                                    serviceState    = d_serviceState,
                                    logs            = str_logs
                                    )
        # d_ret {
        #            'status':      bool,
        #             d_process: {
        #               'currentState':     str_currentState,
        #               'removeJob':        b_removeJob,
        #               'status':           True
        #             }
        #       }

        return {
            'status':           d_ret['status'],
            'logs':             str_logs,
            'currentState':     d_ret['d_process']['currentState']
        }

    def t_delete_process(self,*args, **kwargs):
        """
        Deletes existing jobs. Checks if container environment is OpenShift.
        If yes, call t_delete_process_openshift to delete the job.
        Pending implementation for other container environment.
        """
        status = jid = ''
        if self.container_env == 'openshift':
            self.dp.qprint('Processing openshift....')
            try:
                d_containerStatus    =   self.t_delete_process_openshift(*args, **kwargs)
                status               =   d_containerStatus['status']
                jid                  =   d_containerStatus['jid']
            except Exception as e:
                if e.reason and e.reason == 'Not Found':
                    status = logs = currentState = e.reason
                else:
                    raise e
            d_ret = {
                'action' : 'Delete Job',
                'job_id' : jid,
                'status' : status
            }
            return {
                "d_ret":    d_ret,
                "status":   status
            }


    def t_delete_process_openshift(self,*args, **kwargs):
        """
        Delete job and related resources (pods & pvc) from OpenShift
        """
        jid =  status = None
        for k,v in kwargs.items():
            if k == 'request' and v['action'] == 'delete' :     jid = v['meta']['value']
        d_json  = self.get_openshift_manager().state(jid)
        if d_json['Status'] == 'Not Found':
            status = d_json['Status']
        else:
            self.get_openshift_manager().remove_job(jid)
            self.get_openshift_manager().remove_pvc(jid)
            status = 'Job deleted successfully'
        return {
            "jid" : jid,
            "status" : status
        }

    def t_status_process_openshift(self, *args, **kwargs):
        """
        Determine the status of a job scheduled using the openshift manager.

        PRECONDITIONS:
        o   Only call this method if a container structure exists
            in the relevant job tree!

        POSTCONDITIONS:
        o   If the job is completed, then shutdown the container cluster
            service.
        """
        self.dp.qprint('------- Processing job status within t_status_process_openshift ----------- ')
        str_logs = ''
        # Get job-id from request
        for k,v in kwargs.items():
            if k == 'request' and v['action'] == 'status' :     jid = v['meta']['value']

        # Query OpenShift API to get job state
        d_json  = self.get_openshift_manager().state(jid)

        if d_json['Status']['Message'] == 'finished':
            pod_names = self.get_openshift_manager().get_pod_names_in_job(jid)
            for _, pod_name in enumerate(pod_names):
                str_logs += self.get_openshift_manager().get_job_pod_logs(pod_name, jid)
        else:
            str_logs = d_json['Status']['Message']

        status  = d_json['Status']
        currentState =  d_json['Status']['Message']

        return {
            'status':           status,
            'logs':             str_logs,
            'currentState':     [currentState]
        }

    def t_status_process_openshift_stateObject(self, *args, **kwargs):
        """
        Process the actual JSON container return object on service
        state.

        PRECONDITIONS:
        o   This method should only ever be called by t_status_process_openshift().

        POSTCONDITIONS:
        o   A string denoting the current state is returned.

        """

        def job_exists(jid):
            """
            Returns a bool:
                - True:     <jid> does exist
                - False:    <jid> does not exist
            """
            b_exists        = False
            try:
                job         = self.get_openshift_manager().get_job(jid)
                b_exists    = True
            except:
                b_exists    = False
            return b_exists

        def job_shutDown(d_serviceInfo):
            """
            Shut down a service
            """
            try:
                self.get_openshift_manager().remove_pvc(jid)
                self.get_openshift_manager().remove_job(jid)
            except Exception as err:
                self.dp.qprint("Error deleting pvc/job:", err)

        d_serviceState      = None
        d_jobState          = None
        str_hitIndex        = "0"
        str_logs            = ""
        d_ret               = {}

        for k,v in kwargs.items():
            if k == 'jobState':         d_jobState      = v
            if k == 'serviceState':     d_serviceState  = v
            if k == 'hitIndex':         str_hitIndex    = v
            if k == 'logs':             str_logs        = v
        if d_serviceState:
            d_ret = self.t_status_process_state(**kwargs)
            if d_ret['removeJob']:
                str_jobRoot = d_jobState['d_ret']['%s.container' % (str_hitIndex)]['jobRoot']
                self.within.ptree.cd('/%s' % str_jobRoot)
                jid = self.within.ptree.cat('jid')
                if job_exists(jid):
                    job_shutDown(jid)
        return {
            'status':       True,
            'd_process':    d_ret
        }

    def get_openshift_manager(self):
        if not self.openshiftmgr:
            self.openshiftmgr = OpenShiftManager()
        return self.openshiftmgr

    def t_status_process_state(self, *args, **kwargs):
        """
        This method processes the swarm state object to make the
        final determination on a job's state and print out container
        job state and logs.

        It also returns a signal to the caller to trigger the removal
        of the job from the swarm scheduler if the job has completed.
        """

        def debug_print(    str_jobRoot,
                            d_serviceState,
                            str_currentState,
                            str_logs
                        ):
            """
            Simply print some useful debug info.
            """
            l_commsNorm = ['rx',    'rx',    'tx']
            l_commsErr  = ['error', 'error', 'error']
            l_comms     = l_commsNorm
            if str_currentState == 'finishedWithError':
                l_comms = l_commsErr
            self.dp.qprint('\njobRoot %s\n-->%s<--...' % \
                                    (str_jobRoot,
                                    str_currentState),
                                    comms = l_comms[0])
            self.dp.qprint('\n%s' % self.df_print(d_serviceState),
                                    comms = l_comms[1])
            self.dp.qprint('\njob logs:\n%s' % str_logs,
                                    comms = l_comms[2])

        d_serviceState      = {}
        d_jobState          = {}
        hitIndex            = 0
        str_logs            = ""
        b_status            = False
        str_currentState    = "undefined"

        for k,v in kwargs.items():
            if k == 'jobState':         d_jobState          = v
            if k == 'serviceState':     d_serviceState      = v
            if k == 'hitIndex':         str_hitIndex        = v
            if k == 'logs':             str_logs            = v

        b_removeJob = False
        str_jobRoot = d_jobState['d_ret']['%s.container' % (hitIndex)]['jobRoot']
        str_state   = d_serviceState['Status']['State']
        str_message = d_serviceState['Status']['Message']
        if str_state == 'running' and str_message == 'started':
            str_currentState    = 'started'
            debug_print(str_jobRoot, d_serviceState, str_currentState, str_logs)
            b_status    = True
        else:
            self.DB_store(d_serviceState,   '/%s/container' % (str_jobRoot), 'state')
            self.DB_store(str_logs,         '/%s/container' % (str_jobRoot), 'logs')
            b_removeJob   = True
            if str_state == 'failed'        and str_message == 'started':
                str_currentState    = 'finishedWithError'
                debug_print(str_jobRoot, d_serviceState, str_currentState, str_logs)
            elif str_state == 'complete'    and str_message == 'finished':
                str_currentState    = 'finishedSuccessfully'
                debug_print(str_jobRoot, d_serviceState, str_currentState, str_logs)
            b_status = True
        self.DB_store(str_currentState, '/%s/container' % (str_jobRoot), 'currentState')
        if str_currentState == 'undefined':
            self.dp.qprint('The state of the job is undefined!', comms = 'error')
            self.dp.qprint('This typically means that the scheduler rejected the job.', comms = 'error')
            self.dp.qprint('jobRoot = %s' % str_jobRoot, comms = 'error')
        return {
                    'currentState':     str_currentState,
                    'removeJob':        b_removeJob,
                    'status':           b_status
                }

    def t_hello_process(self, *args, **kwargs):
        """

        The 'hello' action is merely to 'speak' with the server. The server
        can return current date/time, echo back a string, query the startup
        command line args, etc.

        This method is a simple means of checking if the server is "up" and
        running.

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In hello process...")
        b_status            = False
        d_ret               = {}
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta  = d_request['meta']
        if 'askAbout' in d_meta.keys():
            str_askAbout    = d_meta['askAbout']
            d_ret['name']       = self.within.str_name
            d_ret['version']    = self.within.str_version
            if str_askAbout == 'timestamp':
                str_timeStamp   = datetime.datetime.today().strftime('%Y%m%d%H%M%S.%f')
                d_ret['timestamp']              = {}
                d_ret['timestamp']['now']       = str_timeStamp
                b_status                        = True
            if str_askAbout == 'sysinfo':
                d_ret['sysinfo']                = {}
                d_ret['sysinfo']['system']      = platform.system()
                d_ret['sysinfo']['machine']     = platform.machine()
                d_ret['sysinfo']['platform']    = platform.platform()
                d_ret['sysinfo']['uname']       = platform.uname()
                d_ret['sysinfo']['version']     = platform.version()
                d_ret['sysinfo']['memory']      = psutil.virtual_memory()
                d_ret['sysinfo']['cpucount']    = multiprocessing.cpu_count()
                d_ret['sysinfo']['loadavg']     = os.getloadavg()
                d_ret['sysinfo']['cpu_percent'] = psutil.cpu_percent()
                d_ret['sysinfo']['hostname']    = socket.gethostname()
                d_ret['sysinfo']['inet']        = [l for l in ([ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][:1], [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]]) if l][0][0]
                b_status                        = True
            if str_askAbout == 'echoBack':
                d_ret['echoBack']               = {}
                d_ret['echoBack']['msg']        = d_meta['echoBack']
                b_status                        = True

        return { 'd_ret':   d_ret,
                 'status':  b_status}

    def t_run_process(self, *args, **kwargs):
        """
        Main job handler -- this is in turn a thread spawned from the
        parent listener thread.
        By being threaded, the client http caller gets an immediate
        response without needing to wait on the jobs actually running
        to completion.
        """

        str_cmd             = ""
        d_request           = {}
        d_meta              = {}
        d_Tcontainer        = {}

        for k,v in kwargs.items():
            if k == 'request':      d_request       = v
            if k == 'treeList':     d_Tcontainer    = v

        d_meta          = d_request['meta']

        if d_meta:
            self.jid    = d_meta['jid']
            self.auid   = d_meta['auid']
            str_cmd     = d_meta['cmd']

        if isinstance(self.jid, int):
            self.jid    = str(self.jid)

        self.dp.qprint("spawning and starting poller thread")

        # Start the 'poller' worker
        self.poller  = Poller(cmd           = str_cmd,
                              debugToFile   = self.b_debugToFile,
                              verbosity     = self.verbosity,
                              debugFile     = self.str_debugFile)
        self.poller.start()

        str_timeStamp       = datetime.datetime.today().strftime('%Y%m%d%H%M%S.%f')
        str_uuid            = uuid.uuid4()
        str_dir             = '%s_%s' % (str_timeStamp, str_uuid)
        self.str_jobRootDir = str_dir

        b_jobsAllDone       = False
        p                   = self.within.ptree

        p.cd('/')
        p.mkcd(str_dir)

        if d_Tcontainer:
            # Save the trees in this list to the DB...
            for name,tree in d_Tcontainer.items():
                p.mkcd(name)
                tree.copy(startPath = '/', destination = p, pathDiskRoot = '/%s/%s' % (str_dir, name))
                p.cd('/%s' % str_dir)

        p.touch('d_meta',       json.dumps(d_meta))
        for detailKey in ['cmdMgr', 'cmdMgr_byte_str']:
            if detailKey in d_meta.keys():
                p.touch(detailKey,   json.dumps(d_meta[detailKey]))

        p.touch('cmd',          str_cmd)
        if len(self.auid):
            p.touch('auid',     self.auid)
        if len(self.jid):
            p.touch('jid',      self.jid)

        p.mkdir('start')
        p.mkdir('end')

        jobCount        = 0
        p.touch('jobCount',     jobCount)

        while not b_jobsAllDone:
            try:
                b_jobsAllDone   = self.poller.queueAllDone.get_nowait()
            except queue.Empty:
                self.dp.qprint('Waiting on start job info')
                d_startInfo     = self.poller.queueStart.get()
                str_startDir    = '/%s/start/%d' % (self.str_jobRootDir, jobCount)
                p.mkdir(str_startDir)
                p.cd(str_startDir)
                p.touch('startInfo', d_startInfo.copy())
                p.touch('/%s/startInfo' % str_dir, d_startInfo.copy())

                self.dp.qprint('Waiting on end job info')
                d_endInfo       = self.poller.queueEnd.get()
                str_endDir      = '/%s/end/%d' % (self.str_jobRootDir, jobCount)
                p.mkdir(str_endDir)
                p.cd(str_endDir)
                p.touch('endInfo', d_endInfo.copy())
                p.touch('/%s/endInfo' % str_dir,    d_endInfo.copy())

                p.touch('/%s/jobCount' % str_dir,   jobCount)
                jobCount        += 1
        self.dp.qprint('All jobs processed.')

        # Save DB state...
        self.within.ptree           = p
        self.within.DB_fileIO(cmd   = 'save')

    def FScomponent_pollExists(self, *args, **kwargs):
        """
        This method polls access to a file system component (a file or
        directory). Its purpose is to wait for possible transients when
        an asynchronous process creates a file system component that some
        method in pmans wants to access.
        """
        maxLoopTries    = 20
        currentLoop     = 1
        str_dir         = ''

        for k, v in kwargs.items():
            if k == 'maxLoopTries': maxLoopTries    = v
            if k == 'dir':          str_dir         = v

        b_exists        = False
        b_checkAgain    = True
        while b_checkAgain:
            self.dp.qprint('Checking if %s exists (currentLoop: %d)...' % (str_dir, currentLoop), comms = 'rx')
            b_exists    = os.path.exists(str_dir)
            if b_exists:
                b_checkAgain    = False
                self.dp.qprint('Dir exists!', comms = 'rx')
            else:
                self.dp.qprint('Dir does not exist! Sleeping...', comms = 'error')
                time.sleep(2)
            currentLoop += 1
            if currentLoop == maxLoopTries:
                b_checkAgain = False
        return b_exists

    def t_run_process_container(self, *args, **kwargs):
        """
        A threaded run method specialized to handling containerized managers and targets.

        NOTE: If 'serviceName' is not specified/present, then this defaults to the 'jid'
        value and is in fact the default behaviour.

        Typical JSON d_request:

        {   "action": "run",
            "meta":  {
                "cmd":      "$execshell $selfpath/$selfexec --prefix test- --sleepLength 0 /share/incoming /share/outgoing",
                "auid":     "rudolphpienaar",
                "jid":      "simpledsapp-1",
                "threaded": true,
                "container":   {
                        "target": {
                            "image":        "fnndsc/pl-simpledsapp"
                        },
                        "manager": {
                            "image":        "fnndsc/swarm",
                            "app":          "swarm.py",
                            "env":  {
                                "shareDir":     "/home/tmp/share",
                                "serviceType":  "docker",
                                "serviceName":  "testService"
                            }
                        }
                }
            }
        }

        """

        str_cmd             = ""
        str_shareDir        = ""
        str_serviceName     = ""
        d_request           = {}
        d_meta              = {}
        d_container         = {}
        d_image             = {}
        d_manager           = {}
        d_env               = {}

        self.dp.qprint('Processing swarm-type job...')
        for k,v in kwargs.items():
            if k == 'request': d_request    = v

        d_meta          = d_request['meta']

        if d_meta:
            self.jid            = d_meta['jid']
            self.auid           = d_meta['auid']
            str_cmd             = d_meta['cmd']
            str_serviceName     = self.jid

            if 'container' in d_meta.keys():
                d_container                 = d_meta['container']
                d_target                    = d_container['target']
                str_targetImage             = d_target['image']

                d_manager                   = d_container['manager']
                str_managerImage            = d_manager['image']
                str_managerApp              = d_manager['app']

                d_env                       = d_manager['env']
                if 'shareDir' in d_env.keys():
                    str_shareDir            = d_env['shareDir']
                    # Remove trailing '/' if it exists in shareDir
                    str_shareDir            = str_shareDir.rstrip('/')
                    b_exists                = self.FScomponent_pollExists(dir = str_shareDir)
                    if not b_exists:
                        self.dp.qprint('Could not access volume mapped share dir: %s' % str_shareDir, comms = 'error')

                if 'STOREBASE' in os.environ:
                    str_storeBase           = os.environ['STOREBASE']
                    (str_origBase, str_key) = os.path.split(str_shareDir)
                    self.dp.qprint('Overriding shareDir (orig): %s' % str_shareDir)
                    str_shareDir            = os.path.join(str_storeBase, str_key)
                    self.dp.qprint('Overriding shareDir (new):  %s' % str_shareDir)
                if 'serviceName' in d_env.keys():
                    str_serviceName         = d_env['serviceName']
                else:
                    d_env['serviceName']    = str_serviceName

            # First, attach to the docker daemon...
            client = docker.from_env()

            str_cmdLine     = str_cmd
            str_cmdManager  = '%s -s %s -m %s -i %s -p none -c "%s"' % \
                              (str_managerApp, str_serviceName, str_shareDir, str_targetImage, str_cmdLine)
            try:
                byte_str    = client.containers.run('%s' % str_managerImage,
                                             str_cmdManager,
                                             volumes = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}},
                                             remove  = True)
            except Exception as e:
                # An exception here most likely occurs due to a serviceName collision.
                # Solution is to stop the service and retry.
                str_e   = '%s' % e
                print(str_e)

            d_meta['cmdMgr']            = '%s %s' % (str_managerImage, str_cmdManager)
            d_meta['cmdMrg_byte_str']   = str(byte_str, 'utf-8')

            # Call the "parent" method -- reset the cmdLine to an "echo"
            # and create an stree off the 'container' dictionary to store
            # in the pman DB entry.
            d_meta['cmd']   = 'echo "%s"' % str_cmd
            T_container     = C_stree()
            T_container.initFromDict(d_container)
            d_Tcontainer    = {'container': T_container}
            self.t_run_process(request  = d_request,
                               treeList = d_Tcontainer)
            self.dp.qprint('Returning from swarm-type job...')

    def t_run_process_openshift(self, *args, **kwargs):
        """
        A threaded run method specialized for handling openshift
        """

        str_cmd             = ""
        d_request           = {}
        d_meta              = {}
        d_container         = {}
        d_image             = {}

        self.dp.qprint('Processing openshift job...')

        for k,v in kwargs.items():
            if k == 'request': d_request    = v

        d_meta          = d_request['meta']

        if d_meta:
            self.jid            = d_meta['jid']
            self.auid           = d_meta['auid']
            str_cmd             = d_meta['cmd']

            # TODO: Currently it is assumed that incoming and outgoing dir will always be the last two arg
            # It holds true for 'ds' plugins but not for 'fs' plugins.
            # Implementation to support both fs ad ds plugins should be incorporated.
            str_arr = str_cmd.split()
            incoming_dir = str_arr[len(str_arr)-2]
            outgoing_dir = str_arr[len(str_arr)-1]

            if 'number_of_workers' in d_meta:
                number_of_workers = d_meta['number_of_workers']
            else:
                number_of_workers = '1'
            if 'cpu_limit' in d_meta:
                cpu_limit = d_meta['cpu_limit']
            else:
                cpu_limit = '2000m'
            if 'memory_limit' in d_meta:
                memory_limit = d_meta['memory_limit']
            else:
                memory_limit = '1024Mi'

            if 'gpu_limit' in d_meta:
                gpu_limit = d_meta['gpu_limit']
            else:
                gpu_limit = 0

            if 'container' in d_meta:
                d_container                 = d_meta['container']
                d_target                    = d_container['target']
                str_targetImage             = d_target['image']

            # Create the Persistent Volume Claim
            if os.environ.get('STORAGE_TYPE') == 'swift':
                self.dp.qprint("Creating shared PVC")
                self.get_openshift_manager().create_pvc(self.jid)

            str_cmdLine = str_cmd
            self.get_openshift_manager().schedule(str_targetImage, str_cmdLine, self.jid,
                                                  number_of_workers, cpu_limit, memory_limit, gpu_limit,
                                                  incoming_dir, outgoing_dir)

            self.dp.qprint('Returning from openshift job...')

    def json_filePart_get(self, **kwargs):
        """
        If the requested path is *within* a json "file" on the
        DB, then we need to find the file, and map the relevant
        path to components in that file.
        """

    def DB_get(self, **kwargs):
        """
        Returns part of the DB tree based on path spec in the URL
        """

        r           = C_stree()
        p           = self.within.ptree

        pcwd        = p.cwd()
        str_URLpath = "/api/v1/"
        for k,v in kwargs.items():
            if k == 'path':     str_URLpath = v

        str_path    = '/' + '/'.join(str_URLpath.split('/')[3:])

        self.dp.qprint("path = %s" % str_path)

        if str_path == '/':
            # If root node, only return list of jobs
            l_rootdir = p.lstr_lsnode(str_path)
            r.mknode(l_rootdir)
        else:
            # Here is a hidden behaviour. If the 'root' dir starts
            # with an underscore, then replace that component of
            # the path with the actual name in list order.
            # This is simply a short hand way to access indexed
            # offsets.

            l_path  = str_path.split('/')
            jobID   = l_path[1]
            # Does the jobID start with an underscore?
            if jobID[0] == '_':
                jobOffset   = jobID[1:]
                l_rootdir   = list(p.lstr_lsnode('/'))
                self.dp.qprint('jobOffset = %s' % jobOffset)
                self.dp.qprint(l_rootdir)
                try:
                    actualJob   = l_rootdir[int(jobOffset)]
                except:
                    return False
                l_path[1]   = actualJob
                str_path    = '/'.join(l_path)

            r.mkdir(str_path)
            r.cd(str_path)
            r.cd('../')
            # if not r.graft(p, str_path):
            if not p.copy(startPath = str_path, destination = r)['status']:
                # We are probably trying to access a file...
                # First, remove the erroneous path in the return DB
                r.rm(str_path)

                # Now, we need to find the "file", parse the json layer
                # and save...
                n                   = 0
                contents            = p.cat(str_path)
                str_pathFile        = str_path
                l_path              = str_path.split('/')
                totalPathLen        = len(l_path)
                l_pathFile          = []
                while not contents and -1*n < totalPathLen:
                    n               -= 1
                    str_pathFile    = '/'.join(str_path.split('/')[0:n])
                    contents        = p.cat(str_pathFile)
                    l_pathFile.append(l_path[n])

                if contents and n<0:
                    l_pathFile      = l_pathFile[::-1]
                    str_access      = ""
                    for l in l_pathFile:
                        str_access += "['%s']" % l
                    self.dp.qprint('str_access = %s' % str_access)
                    try:
                        contents        = eval('contents%s' % str_access)
                    except:
                        contents        = False

                r.touch(str_path, contents)

        p.cd(pcwd)

        self.dp.qprint(r)
        # self.dp.qprint(dict(r.snode_root))
        self.dp.qprint(self.pp.pformat(dict(r.snode_root)).strip())
        return dict(r.snode_root)

        # return r

    def process(self, request, **kwargs):
        """ Process the message from remote client

        In some philosophical respects, this process() method in fact implements
        REST-like API of its own.

        """
        if len(request):
            REST_header     = ""
            REST_verb       = ""
            str_path        = ""
            json_payload    = ""

            self.dp.qprint("Listener ID - %s: process() - handling request" % (self.worker_id))

            now             = datetime.datetime.today()
            str_timeStamp   = now.strftime('%Y-%m-%d %H:%M:%S.%f')
            self.dp.qprint(Colors.YELLOW)
            self.dp.qprint("***********************************************")
            self.dp.qprint("***********************************************")
            self.dp.qprint("%s incoming data stream" % (str_timeStamp) )
            self.dp.qprint("***********************************************")
            self.dp.qprint("len = %d" % len(request))
            self.dp.qprint("***********************************************")
            self.dp.qprint(Colors.CYAN + "%s\n" % (request.decode()) + Colors.YELLOW)
            self.dp.qprint("***********************************************" + Colors.NO_COLOUR)
            l_raw           = request.decode().split('\n')
            FORMtype        = l_raw[0].split('/')[0]

            self.dp.qprint('Request = ...')
            self.dp.qprint(l_raw)
            REST_header             = l_raw[0]
            REST_verb               = REST_header.split()[0]
            str_path                = REST_header.split()[1]
            json_payload            = l_raw[-1]

            # remove trailing '/' if any on path
            if str_path[-1]         == '/': str_path = str_path[0:-1]

            d_ret                   = {'status': False,
                                       'RESTheader': REST_header,
                                       'RESTverb': REST_verb,
                                       'action': "",
                                       'path': str_path,
                                       'receivedByServer': l_raw}

            self.dp.qprint("Using token authentication: %s" % self.b_tokenAuth)
            if (not self.b_tokenAuth) or self.authModule.authorizeClientRequest(request.decode())[0]:
                self.dp.qprint("Request authorized")
                if REST_verb == 'GET':
                    d_ret['GET']    = self.DB_get(path = str_path)
                    d_ret['status'] = True
                self.dp.qprint('json_payload = %s' % self.pp.pformat(json_payload).strip())
                d_ret['client_json_payload']    = json_payload
                d_ret['client_json_len']        = len(json_payload)
                if len(json_payload):
                    d_payload           = json.loads(json_payload)
                    d_request           = d_payload['payload']
                    payload_verb        = d_request['action']
                    if 'meta' in d_request.keys():
                        d_meta          = d_request['meta']
                    d_ret['payloadsize']= len(json_payload)

                    if payload_verb == 'quit':
                        self.dp.qprint('Shutting down server...')
                        d_ret['status'] = True

                    if payload_verb == 'run' and REST_verb == 'PUT':
                        d_ret['action']     = payload_verb
                        self.processPUT(    request     = d_request)
                        d_ret['status'] = True

                    if REST_verb == 'POST':
                        self.processPOST(   request = d_request,
                                            ret     = d_ret)
            else:
                self.dp.qprint("Request unauthorized")
            return d_ret
        else:
            return False

    def methodName_parse(self, **kwargs):
        """
        Construct the processing method name (string) by parsing the
        d_meta dictionary.
        """
        d_meta              = {}
        str_method          = ""        # The main 'parent' method
        str_methodSuffix    = ""        # A possible 'subclass' specialization

        for k,v in kwargs.items():
            if k == 'request': d_request= v
        payload_verb        = d_request['action']

        if 'meta' in d_request.keys():
            d_meta          = d_request['meta']

        if 'container' in d_meta.keys():
            if self.container_env == 'openshift':
                # append suffix _openshift to redirect to openshift function
                str_methodSuffix    = '_openshift'
            elif self.container_env == 'swarm':
                # append suffix _container to redirect to container function
                str_methodSuffix    = '_container'

        str_method  = 't_%s_process%s' % (payload_verb, str_methodSuffix)
        return str_method

    def processPOST(self, **kwargs):
        """
         Dispatcher for POST
        """

        for k,v in kwargs.items():
            if k == 'request':  d_request   = v
            if k == 'ret':      d_ret       = v

        payload_verb        = d_request['action']
        if 'meta' in d_request.keys():
            d_meta          = d_request['meta']

        d_ret['action'] = payload_verb
        d_ret['meta']   = d_meta

        b_threaded      = False
        if 'threaded' in d_meta.keys():
            b_threaded  = d_meta['threaded']

        if b_threaded:
            self.dp.qprint("Will process request in new thread.")
            pf_method   = None
            str_method  = self.methodName_parse(request = d_request)
            # str_method  = 't_%s_process' % payload_verb
            try:
                pf_method  = getattr(self, str_method)
            except AttributeError:
                raise NotImplementedError("Class `{}` does not implement `{}`".format(pman.__class__.__name__, str_method))

            t_process           = threading.Thread(     target      = pf_method,
                                                        args        = (),
                                                        kwargs      = kwargs)
            t_process.start()
            time.sleep(0.1)
            # if payload_verb == 'run':
            #     d_ret['jobRootDir'] = self.str_jobRootDir
            d_ret['status']     = True
        else:
            self.dp.qprint("Will process request in current thread.")
            d_done              = eval("self.t_%s_process(request = d_request)" % payload_verb)
            try:
                d_ret['d_ret']      = d_done["d_ret"]
                d_ret['status']     = d_done["status"]
            except:
                self.dp.qprint("An error occurred in reading ret structure. Should this method have been threaded?")

        return d_ret

    def processPUT(self, **kwargs):
        """
         Dispatcher for PUT
        """

        d_request       = {}
        str_action      = "run"
        str_cmd         = "save"
        str_DBpath      = self.str_DBpath
        str_fileio      = "json"
        tree_DB         = self.within.ptree

        for k,v in kwargs.items():
            if k == 'request':  d_request   = v

        str_action      = d_request['action']
        self.dp.qprint('action = %s' % str_action)
        d_meta              = d_request['meta']
        self.dp.qprint('action = %s' % str_action)

        # Optional search criteria
        if 'key'        in d_meta:
            d_search    = self.t_search_process(request = d_request)['d_ret']

            Tj          = C_stree()
            Tdb         = C_stree()
            for j in d_search.keys():
                d_j = d_search[j]
                for job in d_j.keys():
                    str_pathJob         = '/api/v1/' + job

                    d_job               = self.DB_get(path = str_pathJob)
                    Tj.initFromDict(d_job)
                    Tj.copy(startPath = '/', destination = Tdb)

                    # Tdb.graft(Tj, '/')

                    # self.DB_get(path = str_pathJob).copy(startPath = '/', destination = Tdb)


            # print(Tdb)
            tree_DB     = Tdb

        if 'context'    in d_meta:  str_context     = d_meta['context']
        if 'operation'  in d_meta:  str_cmd         = d_meta['operation']
        if 'dbpath'     in d_meta:  str_DBpath      = d_meta['dbpath']
        if 'fileio'     in d_meta:  str_fileio      = d_meta['fileio']

        if str_action.lower() == 'run' and str_context.lower() == 'db':
            self.within.DB_fileIO(  cmd         = str_cmd,
                                    fileio      = str_fileio,
                                    dbpath      = str_DBpath,
                                    db          = tree_DB)

class Poller(threading.Thread):
    """
    The Poller checks for running processes based on the internal
    DB and system process table. Jobs that are no longer running are
    removed from the internal DB.
    """

    def __init__(self, **kwargs):

        self.pollTime           = 10
        self.str_cmd            = ""
        self.crunner            = None
        self.queueStart         = queue.Queue()
        self.queueEnd           = queue.Queue()
        self.queueAllDone       = queue.Queue()
        self.__name__           = 'Poller'

        # self.dp.qprint('starting...', level=-1)

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.verbosity          = 1

        for key,val in kwargs.items():
            if key == 'pollTime':       self.pollTime       = val
            if key == 'cmd':            self.str_cmd        = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val
            if key == 'verbosity':      self.verbosity      = int(val)

        self.dp                 = pfmisc.debug(
                                        verbosity   = self.verbosity,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile,
                                        within      = self.__name__)

        threading.Thread.__init__(self)


    def run(self):

        timeout = 1
        loop    = 10

        """ Main execution. """

        # Spawn the crunner object container
        self.crunner  = Crunner(cmd         = self.str_cmd,
                                debugToFile = self.b_debugToFile,
                                verbosity   = self.verbosity,
                                debugFile   = self.str_debugFile)
        self.crunner.start()

        b_jobsAllDone   = False

        while not b_jobsAllDone:
            try:
                b_jobsAllDone = self.crunner.queueAllDone.get_nowait()
            except queue.Empty:
                # We basically propagate the queue contents "up" the chain.
                self.dp.qprint('Waiting on start job info')
                self.queueStart.put(self.crunner.queueStart.get())

                self.dp.qprint('Waiting on end job info')
                self.queueEnd.put(self.crunner.queueEnd.get())

        self.queueAllDone.put(b_jobsAllDone)
        self.dp.qprint("done with Poller.run")

class Crunner(threading.Thread):
    """
    The wrapper thread about the actual process.
    """

    def __init__(self, **kwargs):
        self.__name             = "Crunner"

        self.queueStart         = queue.Queue()
        self.queueEnd           = queue.Queue()
        self.queueAllDone       = queue.Queue()

        self.str_cmd            = ""

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True
        self.verbosity          = 1

        for k,v in kwargs.items():
            if k == 'cmd':          self.str_cmd        = v
            if k == 'debugFile':    self.str_debugFile  = v
            if k == 'debugToFile':  self.b_debugToFile  = v
            if k == 'verbosity':    self.verbosity      = int(v)

        self.shell              = crunner(  verbosity   = self.verbosity,
                                            debugToFile = self.b_debugToFile,
                                            debugFile   = self.str_debugFile)

        self.dp                 = pfmisc.debug(
                                            verbosity   = self.verbosity,
                                            debugFile   = self.str_debugFile,
                                            debugToFile = self.b_debugToFile,
                                            within      = self.__name)
        self.dp.qprint('starting crunner...')

        threading.Thread.__init__(self)

    def jsonJobInfo_queuePut(self, **kwargs):
        """
        Get and return the job dictionary as a json string.
        """

        str_queue   = 'startQueue'
        for k,v in kwargs.items():
            if k == 'queue':    str_queue   = v

        if str_queue == 'startQueue':   queue   = self.queueStart
        if str_queue == 'endQueue':     queue   = self.queueEnd

        # self.dp.qprint(self.shell.d_job)

        queue.put(self.shell.d_job.copy())

    def run(self):

        timeout = 1
        loop    = 10

        """ Main execution. """
        self.dp.qprint("running...")
        self.shell(self.str_cmd)
        # self.shell.jobs_loopctl(    onJobStart  = 'self.jsonJobInfo_queuePut(queue="startQueue")',
        #                             onJobDone   = 'self.jsonJobInfo_queuePut(queue="endQueue")')
        self.shell.jobs_loopctl(    onJobStart  = partial(self.jsonJobInfo_queuePut, queue="startQueue"),
                                    onJobDone   = partial(self.jsonJobInfo_queuePut, queue="endQueue"))
        self.queueAllDone.put(True)
        self.queueStart.put({'allJobsStarted': True})
        self.queueEnd.put({'allJobsDone': True})
        # self.shell.exitOnDone()
        self.dp.qprint('Crunner.run() returning...')
